/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* phone interface subs for z8 (S A Uhler) */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sgtty.h>
#include <stdio.h>

#define usleep(x) \
   { \
   struct timeval time; \
   time.tv_sec = 0; \
   time.tv_usec = (x); \
   select(0,0,0,0,&time); \
   }

struct timeval poll = {0,0};

#define WAIT			250		/* 100ths-sec to wait for write to port */
#define dprintf		if (debug) fprintf
#define dputc			if (debug) putc

/* open the phone port */

static int timeout = 0;			/* true if we had a timeout */
static int alarm_catch();
static char line[512];			/* big enough for anything (yea right!) */
char *index();
int debug;

int
open_phone(port,how)
char *port;
int how;
	{
	int fd;
	int fast = getenv("FAST");

	debug = getenv("DEBUG");

	if (how)
		fd = open(port,O_RDWR | O_NDELAY,0);
	else
		fd = open(port,O_RDWR,0);
	
	if (fd <0)  {
		dprintf(stderr,"  open_phone: Can't open %s\n",port);
		return(-1);
		}

	if (ioctl(fd,TIOCEXCL) != 0) {
		dprintf(stderr,"  open_phone: Can't set %s to exclusive use\n",port);
		return(0);
		}
	set_mode(fd,RAW,ECHO,fast?B38400:B9600);
	return(fd);
	}

/* see if program is running */

int
check_port(fd)
int fd;
	{
	int n;

	write(fd,"?",1);
	dprintf(stderr,"  checking port\n");
	n = tread(fd,line,'\n',WAIT);
	if (n>0 && (line[0]&0x7f) == 'A') {		/* OK */
		dprintf(stderr,"  Phone state: %.3s\n",line);
		return(1);
		}
	else {
		write(fd,"\033",1);	/* cancel */
		dprintf(stderr,"  %c != %c\n",line[0],'A');
		dprintf(stderr,"  Phone interface program is down\n");
		return(0);
		}
	}

/* go off hook (and turn off DCD) */

int
off_hook(fd)
int fd;
	{
	int n;

	write(fd,"H",1);
	n = tread(fd,line,'\n',WAIT);
	if (n > 0 && (*line&0x7f) == 'Y') {		/* OK */
		dprintf(stderr,"  Phone went OFF hook\n");
		return(1);
		}
	else if (n > 0 && (*line&0x7f) == 'N') {		/* OK */
		dprintf(stderr,"  Phone ALREADY OFF hook\n");
		return(1);
		}
	else {
		dprintf(stderr,"  Phone didn't go OFF hook, got: %s\n",line);
		return(0);
		}
	}

/* go on hook */

int
on_hook(fd)
int fd;
	{
	int n;

	write(fd,"h",1);
	n = tread(fd,line,'\n',WAIT);
	if (n > 0 && (*line&0x7f) == 'Y') {		/* OK */
		dprintf(stderr,"  Phone went ON hook\n");
		dcd_off(fd);
		return(1);
		}
	else if (n > 0 && (*line&0x7f) == 'N') {		/* OK */
		dprintf(stderr,"  Phone ALREADY ON hook\n");
		return(1);
		}
	else {
		dprintf(stderr,"  Phone didn't go ON hook, got: [%s]\n",line);
		return(0);
		}
	}

/* disconnect (i.e. turn DCD off) */

int
dcd_off(fd)
int fd;
	{
	int n;

	write(fd,"d",1);
	n = tread(fd,line,'\n',WAIT);
	if (n > 0 && (*line&0x7f) == 'Y') {		/* OK */
		dprintf(stderr,"  DCD turned off\n");
		return(1);
		}
	else if (n > 0 && (*line&0x7f) == 'N') {		/* OK */
		dprintf(stderr,"  DCD ALREADY off\n");
		return(1);
		}
	else {
		dprintf(stderr,"  DCD unknown state, got: [%s]\n",line);
		return(0);
		}
	}

/* interrupt the z8 controller */

int
intr(fd)
int fd;
	{
	write(fd,"\033",1);
	return(tread(fd,line,':',WAIT));
	}

int
set_mode(file,on,off,speed)
int file;      /* file pointer */
int on;        /* flags to turn on */
int off;    	/* flags to turn off */
int speed;		/* set the baud rate (if !0) */
	{
   struct sgttyb buff;

   if (gtty(file,&buff) != 0)
		perror("getty");
   buff.sg_flags |= on;
   buff.sg_flags &= ~off;
   if (speed)
      buff.sg_ispeed = buff.sg_ospeed = speed;
   if (stty(file,&buff) != 0)
		perror("stty");
        return(0);
	}

/* read until a particular character is recieved, or timeout */

static struct itimerval val = {0, 0, 0, 0};

int
tread(fd,data,c,tm)
int fd;		/* file to read from */
char *data;	/* data to read */
int tm;		/* max time to wait (in 100ths seconds) */
char c;		/* read till what char */
	{
	int n;
	void (*old)();

	val.it_value.tv_sec= tm/100;
	val.it_value.tv_usec= (tm%100) * 10000;

	old = signal(SIGALRM,alarm_catch);
	siginterrupt(SIGALRM,1);
	timeout = 0;
	setitimer(ITIMER_REAL,&val,0);
	n = read_till(fd,data,c);
	alarm(0);
	signal(SIGALRM,old);
	return(timeout ? 0 : n);
	}

/* catch the alarm timer */

static int
alarm_catch()
	{
	timeout = 1;
	dprintf(stderr,"  phone_subs: timeout\n");
	return(0);
	}

/* read till timeout */

int
Read(fd,data,max,tm)
int fd;     /* file to read from */
char *data; /* data to read */
int max;
int tm;     /* max time to wait (in 100ths seconds) */
   {
   int n;
   void (*old)();

   val.it_value.tv_sec= tm/100;
   val.it_value.tv_usec= (tm%100) * 10000;

   old = signal(SIGALRM,alarm_catch);
   siginterrupt(SIGALRM,1);
   timeout = 0;
   setitimer(ITIMER_REAL,&val,0);
   n = read(fd,data,max);
   alarm(0);
   signal(SIGALRM,old);
   return(timeout ? 0 : n);
   }

/* read characters till I get what I want */

int
read_till(fd,buff,c)
int fd;		/* what to read */
char *buff;	/* where to put it */
char c;		/* what to read to */
	{
	register int count = 0;

	dprintf(stderr,"Reading: ");
	while(read(fd,buff,1)>0 && (*buff&= 0x7F,*buff++ != c)) {
		dputc(buff[-1],stderr);
		count++;
		}
	dputc('\n',stderr);
	return(count);
	}

/* download a line to the z8, a character at a time */

int
put_line(buff,fd,del)
char *buff;		/* line to put (\n terminated) */
int fd;			/* file descriptor to put */
int del;			/* true if swallow prompt */
	{
	register char *p;
	char c;
	int ln;

	timeout = 0;
	if (!debug) {
		ln = atoi(buff);
		fprintf(stderr,"\r%d  ",ln);
		}
	alarm(6);		/* max time allowed to download a line */
	p = buff;
	while (!timeout && !(*p == '\n' || *p == '\0')) {
		write(fd,p,1);
		if (debug)
			usleep(8000);
		read(fd,&c,1);
		dputc(c,stderr);
		p++;
		}

	/* finish the line */

	write(fd,"\r",1);
	if (debug)
		usleep(40000);
	if (del) {
		int n = read(fd,line,100);
		if (debug)
			write(2,line,n);
		}
	/* dprintf(stderr,"line %s%s",line,*p=='\0'?"\n":""); */

	alarm(0);
	if (timeout) {
		timeout = 0;
		return(1);
		}
	else
		return(0);
	}

#define is_white(c)		(c==' ' || c=='\t' || c=='\n')

int
load(fd,s,n1,n2)
int fd;				/* port */
char *s;				/* file to load */
int n1,n2;			/* line range (if specified) */
	{
	char cnvt[3];
	FILE *in;
	char *p;
	char *rindex();
	int start=0, end=99999;
	int n;
	int pack = (int) getenv("PACK");		/* pack the program */
	int tty=isatty(fd);

	fprintf(stderr,"\r\nLoading file %s\r\n",s);

	if (n1 > 0)
		start=n1;

	if (n2 > n1)
		end = n2;

	if ((in = fopen(s,"r")) == NULL) {
		fprintf(stderr,"Can't open %s\r\n",s);
		return(0);
		}

#ifndef NOCPP
	/* invoke cpp */

	fclose(in);
	sprintf(line,"/usr/lib/cpp -P -I. < %s",s);
	in = popen(line,"r");	

	if (!in) {
		fprintf(stderr,"Sorry, can't find cpp\n");
		return(0);
		}

#endif

	/* download the file */

	while(fgets(line,sizeof(line),in)) {	/* fetch a line */
		int reads = 1;

		/* check for interrupt */

		if (select(32,&reads,0,0,&poll)==1) {
			fprintf(stderr,"\rInterrupted\r\n");
			break;
			}
	
		/* toss blank lines and comments */

		if (*line == '#') {
			fprintf(stderr,"\r\n%s\r",line);
			continue;
			}

		/* kill trailing white space */

		p = line + strlen(line)-1;
		while(p>line && is_white(*p))
			p--;
		*++p = '\n';

		if (*line == '\n')
			continue;

		/* out of range */

		if ((n=atoi(line))>0 && start>0 && (n < start || n > end))
			continue;

		/* convert 'x' to %xx */

		p = line;
		while(*p) {
			if (p[0] == '\'' && p[2]=='\'') {
				sprintf(cnvt,"%%%.2x",p[1]);
				strncpy(p,cnvt,3);
				p+=3;
				}
			else
				p++;
			}

		/* pack it */

		if (pack) {		/* zap all white space */
			int quote=0;		/* in a quoted string */
			char *src = line;	/* src posisiton in line */
			char *dst = line; /* dst position in line */

			while(*src) {
				if (*src=='"') quote = 1-quote;
				if (quote || *src != ' ')
					*dst++ = *src;
				src++;
				}
				*dst = '\n';
			}

		/* output the line, a char at a time */

		if (tty)
			put_line(line,fd,1);
		else
			write(fd,line,index(line,'\n')-line+1);
		}
#ifndef NOCPP
	pclose(in);
#else
	fclose(in);
#endif
	dprintf(stderr,"\r\nDone with download\r\n");
	return(1);
	}
