/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* phone interface, version 5 (use 2090 and ch1817) using BCC11 (or equiv).
 *  - rescued from 2/27/92 version
 *  - Change CH1810 to CH1917 - ring detect is different
 *  - use ST2090 for:
 *    touch tone detection (same as ST204)
 *    touch tone generation (codes are inverted)
 *    call progress monitoring
 *  - ENABLE is now tone detection/generation switch
 *
 * phone interface, version 4 
 *  - pre-process with cpp -P
 *  - remove trailing white space on every line
 *  - convert all 'X' to %xx
 *  - Use touch tone get/send chip
 *
 * Version 3 is too slow, misses touch tones
 *
 * Rules:
 *  + you get variables [a-z]
 *  + Goto or Gosub MUST be the last command on a line
 *  + Basic searches from line 1 on each goto or gosub
 *     -so- goto targets near the top run faster.
 *  (i.e. Its not my fault).
 *
 */

/*
 *   Time critical loop: touch-tone / hangup detection while off-hook
 */

/* Z8 port assignments:
 *
 *  P3/1 (in) ring detect		2		(0 is ringing)
 *  P3/2 (in) DTMF valid		4		(1 is valid tone present)
 *  P3/3 (in) !hangup			8		(0 is NO loop current)
 *  P3/4 (out) DCD				16		(0 is DCD asserted)
 *  P3/5 (out) Off Hook			32		(1 is off hook)
 *  P3/6 (out) DTMF enable		64		(0 is detect, 1 is send)
 *
 *  P2/0 - P2/3 (in) touch tone data
 *  P2/0 - P2/3 (out) touch tone data (inverted)
 *  P2/4 - relay 1 (out)
 *  P2/5 - relay 2 (out)
 *  P2/6 - led (out)
 *  P2/7 - call progress band active detection (in)
 */

/* port 3 assignments */

#define RING		2		/* (in) !ringing */
#define VALID		4		/* (in) valid touch-tone */
#define NOHANG		8		/* (in) line current dropped out when 0 */
#define DCD			16		/* (out) DCD off when asserted */
#define HOOK		32		/* (out) off hook when asserted */
#define ENABLE		64		/* (out) low = recieve, high - send */
#define OH_OK		12		/* bits to check while off hook */

/* port 2 assignments */

#define TONE_MASK		%F		/* touch tone mask */
#define RLYS			%30	/* audio relays - assert to capture audio */
#define RLY_I			%10	/* input to sparc */
#define RLY_O			%20	/* output from sparc */
#define LED				%40	/* led on when asserted */
#define PROGRESS		%80	/* call progress detect (in) */

/* variable assignments (basic needs single character variable names) */

#define ch		c				/* input character from serial port */
#define dcd		d				/* dcd is enabled */
#define enable	e				/* touch tone decoding is enabled */
#define hook	h				/* hook state, ON/OFF */
#define light	l				/* light counter */
#define on		o				/* when to turn light on */
#define prg		p				/* call progress monitoring */
#define dcon	q				/* send a disconnect */
#define ring	r				/* ringing state 0==ringing */
#define state  s				/* which state to be in: on/off hook (goto target) */
#define sent_r	v				/* already sent ringing */
#define last_p	w				/* previous detect state */
#define p2		x				/* cached value of port2 */
#define p3		y				/* cached value of port3 */
#define blink	b				/* how fast to blink light */

/* misc Z8 assignments */

#define port_3		@3	
#define port_2		@2
#define mode_2		@246		/* port 2 mode register (1-in, 0-out) */
#define mode_3		@247		/* port 3 mode register */

#define ON			0								/* on hook */
#define OFF			1								/* off hook */

/* other constants */

#define HNG_WAIT	75			/* how long to wait after hangup to flip relays */
#define DEBOUNCE	80			/* debounce off-hook x-ition.  (loop count*8) */
#define SWITCH		@%FFFD	/* configuration switch setting */
#define CFG			hex(SWITCH)		/* ascii value of switch */
#define RNG			5			/* # of loops to debounce ringing */
#define TONE		7			/* # loops for tone on */
#define SLOW		15			/* slow blink rate */
#define FAST		4			/* fast blink rate */

/* useful macros */

#define getc()			and(@240,%7F)			/* read a character */
#define putc(x)		@240=x					/* write a character */
#define is_char()		and(@250,8)				/* see if a character is ready */
#define or(a,b)		-1-and(-1-a,-1-b)		/* no built in or? */
#define not(x)			(255-x)					/* byte wise not */

/* initialization */

10 goto 800										/* initialization */

/* off hook state (near top of program), look for TT's, hangup, char, or CP */

20 p3=port_3:p2=port_2						/* cache port values */
25 if and(p3,OH_OK)<>NOHANG goto 300	/* get touch tones or hangup */
30 if is_char()<>0 gosub 500				/* got a char, process */
35 if prg=1 goto 150							/* do call progress detection */
40 goto state									/* state is one of {20,50} */

/* on hook state, wait for ringing or character */

50 p3=port_3:p2=port_2									/* cache port values */
60 if and(RING,p3) <> ring gosub 200				/* ringing state change */
70 if sent_r>0 sent_r=sent_r+1:if sent_r>RNG "r":sent_r=0:light=0:on=blink /* ring off */
80 light=light+1											/* blink the LED */
90 if light=on if and(p2,LED)=0 port_2=p2+LED	/* "" */
100 if light>on light=0:port_2=and(p2,not(LED))	/* "" */
110 if is_char()<>0 gosub 500							/* got a char, process */
120 goto state												/* state is one of {20,50} */

/* call progress detection */

150 i=and(PROGRESS,p2): if last_p=i goto state	/* no change */
160 last_p=i:if last_p=PROGRESS "P": goto state	/* 0->1 transition */
170 "p": goto state										/* 1->0 transition */

/* got ringing - CH1817 version */

200 ring=RING-ring								/* toggle ring state */
210 if sent_r=0 "R" on=1:light=0:if dcd=0 dcd=1:port_3=p3-DCD	/* dcd on */
220 sent_r=1:ret									/* reset counter */

/* check for hangup */

300 if and(p3,NOHANG)=0 goto 370				/* got disconnect signal */

/* check for touch tones */

305 if enable=0 goto state						/* tt's not enabled */
310 t=and(p2,TONE_MASK):i=360					/* fetch tone, set goto target */
320 if t=0 print "D":goto i					/* other symbols */
330 if t<10 print HEX(T):goto i				/* decimal digit */
340 goto 350+t-10

350 "0":goto i
351 "*":goto i
352 "#":goto i
353 "A":goto i
354 "B":goto i
355 "C"	/* fall through */

360 if and(port_3,VALID)=0 goto state		/* wait for tone to end */
365 goto i

/* loop interruption send a 'd', optionally de-assert DCD */

370 if dcon=0 goto state								/* disconnect disabled */
380 "d":state=50:i=0:port_3=p3-HOOK:hook=ON		/* go on hook */
385 if dcd=1 if and(SWITCH,%80)=0 dcd=0 port_3=port_3+DCD	/* de-assert DCD */
390 i=i+1: if i<HNG_WAIT goto 390					/* let audio drain */
395 port_2=0: goto state								/* turn off relays */

/* go off hook */

400 if hook=OFF "N":ret							/* already off hook, fail */
410 if and(RING,port_3)=0 goto 410			/* make sure no ringing */
420 port_3=port_3+HOOK:ring=RING:hook=OFF	/* go off hook */
430 last_p=0:port_2=RLYS+LED:on=blink:state=20:i=0	/* turn on audio */

/* debounce off-hook */

440 i=i+and(port_3,NOHANG):if i>DEBOUNCE  "Y":ret	/* check loop current */
445 goto 440

/* go on hook */

450 if hook=ON "N":ret								/* already on, fail */
460 "Y":sent_r=0:state=50:port_2=0:port_3=p3-HOOK:hook=ON:ret

/* flash hook - called only when off hook */

470 port_3=p3-HOOK:i=0								/* go on hook */
480 i=i+1:if i<20 goto 480							/* wait a bit */
490 port_3=p3:i=0:goto 440							/* off hook and debounce  */

/* handle characters from the port */

500 ch=getc()											/* get the character */
505 if ch='H' goto 400								/* go off-hook */
510 if ch='h' goto 450								/* go on-hook */
515 if ch='d' if dcd=1 "Y":dcd=0:port_3=p3+DCD:ret	/* turn dcd off */
520 if ch='D' if dcd=0 "Y":dcd=1:port_3=p3-DCD:ret	/* turn dcd on */
525 if ch='E' if enable=0 "Y":enable=1:ret	/* TT detect on */
530 if ch='e' if enable=1 "Y":enable=0:ret	/* TT detect off */
535 if ch='F' if hook=OFF goto 470				/* flash hook */
540 if ch='V' "STIM 2.7/";CFG;" (c) 2/26/92 Bellcore (by S A Uhler)":ret
545 if ch='X' "X":ret								/* echo - for sync. */
550 if ch='I' if and(p2,RLYS)=RLYS "Y":port_2=p2-RLY_O:ret	/* bypass input */
555 if ch='O' if and(p2,RLYS)=RLYS "Y":port_2=p2-RLY_I:ret	/* bypass output*/
560 if ch='A' if and(p2,RLYS)<>0:"Y":port_2=and(p2,not(RLYS))+RLYS: ret
565 if ch='L' if dcon=0 dcon=1:"Y":ret			/* enable hangup detection */
570 if ch='l' if dcon=1 dcon=0:"Y":ret			/* disable hangup detection */
575 if ch='P' if prg=0 "Y":last_p=0:prg=1:ret /* progress detect on */
580 if ch='p' if prg=1 "Y":prg=0:ret			/* progress detect off */
582 if ch='C' if blink=SLOW blink=FAST:on=blink:"Y":ret	/* set fast blink */
583 if ch='c' if blink=FAST blink=SLOW:on=blink:"Y":ret	/* set slow blink */
585 if ch='?' goto 600  					      /* print status */
590 if ch='R' "Y":goto 800							/* Reset to startup state */
595 if hook=OFF goto 700							/* check for tone generation */
599 "N" ret

/* print status */

600 ch='A'
610 putc(ch):ch='d':if dcd=1 ch='D'				/* DCD state */
620 putc(ch):ch='h':if hook=1 ch='H'			/* hook state */
630 putc(ch):ch='r':if ring<>RING ch='R'		/* ringing (unreliable) */
640 putc(ch):ch='e':if enable=1 ch='E'			/* TT enabled state */
650 putc(ch):ch='l':if dcon=1 ch='L'			/* loop enabled state */
660 putc(ch):ch='p':if prg=1 ch='P'				/* call progress state */
670 putc(ch):".":ret

/* generate a touch tone  - get code for tone */

700 if ch>'0' if ch <':' i=ch-'0': goto 750	/* 1 - 9 */
710 if ch='*' i=11: goto 750
720 if ch='0' i=10: goto 750
730 if ch='#' i=12: goto 750							
740 "N": ret											/* invalid command */

/* send the tone */

750 mode_2=%80:port_2=and(p2,%F0)+15-i:i=0:port_3=p3+ENABLE /* start tone */
760 i=i+1: if i<TONE goto 760						/* wait */
770 port_3=p3:mode_2=%8F:port_2=p2				/* stop tone */
780 "Y":ret												/* send ack */

/* initialization */

800 prg=0:dcd=0:enable=0:sent_r=0:light=0:dcon=0	/* initialize flags */
810 ring=RING:hook=ON:blink=SLOW:on=blink:state=50	/* initialize state */
820 port_2=0:port_3=DCD										/* set port values */
830 mode_2=%8F:mode_3=%41: goto state					/* configure ports */

900 rem end
