/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* basic z8 communication program with provisions for downloading BASIC (S A Uhler) */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sgtty.h>
#include <stdio.h>

#define PORT		"/dev/call"
#define MSG			"\rInvalid command, ignored\n"
#define MSG2		"\rEnter file to put:     \b\b\b\b\b "
#define MSG3		"\rEnter file to get:     \b\b\b\b\b "
#define MSG4		"\rCommand (g,p,r,d,l,!,.): "
#define MSG5		"\rEnter basic command to start dump: "
#define MSG6		"\rEnter file to store it in: "

#define dprintf	if(debug)fprintf

char line[512];
char buff[512];
char last_file[256];

char *getenv(), *index();
int timeout = 0;	/* true iff timeout on read */
int literal=0;		/* for literal mode */
static int catch();
int debug;

int
main(argc,argv)
int argc;
char **argv;
	{
	char c;			/* current character read */
	int fd;			/* file desc for port */
	char *port=PORT;		/* which port */
	int mask, reads;	/* for select */
	int done=0;
	int n;
	int dump_fd;
	char *s;
	int l1=0, l2=0;

	debug = (int) getenv("DEBUG");
	if (getenv("PORT"))
		port = getenv("PORT");

	if ((fd=open_phone(port,1)) < 0) {
		perror(port);
		exit(1);
		}

	setbuf(stderr,NULL);
	set_mode(0,RAW,ECHO,0);

	/* start a program */

	if (argc > 1) {
		if (intr(fd))
			load(fd,argv[1],0,0);
		else
			fprintf(stderr,"%s: STIM doesn't respond download cancelled\n"
						,*argv);
		}

	/* look for input */

	mask = 1<<0 | 1<<fd;

	printf("Connected\r\n");
	while(!done) {
		reads = mask;
		while((n=select(32,&reads,0,0,0))<0) {
			printf(stderr,"Select error\r\n");
			break;
			}

		if (reads & 1<<0) {	/* got terminal input */
			read(0,&c,1);
			switch(c) {
				case 0x13:			/* x -off */
					fprintf(stderr,"(Stopped..");
					fflush(stderr);
					read(0,&c,1);
					fprintf(stderr,"resumed)");
					fflush(stderr);
					break;
				case '\004':		/* quit */
					done=1;
					fprintf(stderr,"done\r\n");
					break;
				case '~':			/* escape */
					write(1,MSG4,strlen(MSG4));
					read(0,&c,1);
					switch(c) {
						case 'd':			/* dump memory */
							fetch(line,MSG5);
							if (strlen(line) < 2)
								break;
							fetch(buff,MSG6);
							if (!*buff)
								break;
							dump_fd = open(buff,O_WRONLY|O_CREAT,0644);
							put_line(line,fd,0);
							while(n = Read(fd,line,512,200)) {
								write(dump_fd,line,n);
								dprintf(stderr,"Dump got %d\r\n",n);
								if (Read(0,&c,1,100) && c=='\004') {
									fprintf(stderr,"\r\ninterrupted\r\n");
									break;
									}
								}
							fprintf(stderr,"\r\ndone\r\n");
							close(dump_fd);
							break;
						case 'g':			/* get a program */
							fetch(line,MSG3);
							if (*line)
								get(line,fd);
							break;
						case 'r':			/* re-open the port */
							close(fd);
							fd=open_phone(port,1);
							fprintf(stderr,"reset port %s\r\n",fd>0?"OK":"ERROR");
							break;
						case 'P':			/* download a program (to other file) */
						case 'p':			/* download a program */
							fetch(line,MSG2);
							if (!*line)
								break;
							if (c=='P') {
								fetch(buff,MSG6);
								if (!*buff)
									break;
								}
							if (s = index(line,' ')) {
								*s++ = '\0';
								strcpy(last_file,line);
								l1 = atoi(s);
								l2 = (s=index(s,' ')) ? atoi(s+1) : 99999;
								}
							else
								l1=l2=0;
							if (c=='P') {
								dump_fd = open(buff,O_WRONLY|O_CREAT,0666);
								load(dump_fd,line,l1,l2);
								close(dump_fd);
								}
							else
								load(fd,line,l1,l2);
							fprintf(stderr,"\r\ndone\r\n");
							break;
						case 'l':		/* toggle literal mode */
							literal= 1-literal;
							fprintf(stderr,"literal mode %s\r\n",
										literal?"on":"off");
							break;
						case '.':			/* exit */
							done++;
							break;
						case 'e':			/* run vi on last loaded program */
							set_mode(0,ECHO,RAW,0);
							sprintf(line,"%s +%d %s",
								getenv("EDITOR")?getenv("EDITOR"):"vi",
								l1, last_file);
							system(line);
							set_mode(0,RAW,ECHO,0);
							fprintf(stderr,"\r\nResume\r\n");
							break;
						case '!':			/* do a shell */
							set_mode(0,ECHO,RAW,0);
							if (getenv("SHELL"))
								system(getenv("SHELL"));
							else
								system("/bin/csh");
							set_mode(0,RAW,ECHO,0);
							fprintf(stderr,"\r\nResume\r\n");
							break;
						case '~':			/* put a ~ */
							write(fd,&c,1);
							break;
						default:
							write(1,MSG,strlen(MSG));
						}
					break;
				default:
					write(fd,&c,1);
					break;
				}
			}

		if (reads & 1<<fd) {	/* got input from port */
			int n;
			n = read(fd,line,sizeof(line));
			if (literal) {
				char *s = line;
				while (n-- > 0)
					write_literal(1,*s++);
				}
			else
				write(1,line,n);
			}
		}

	set_mode(0,ECHO,RAW,0);
	exit(0);
	}

/* prompt for a line */

int
fetch(line,msg)
char *line;			/* where to put command */
char *msg;		/* prompt message */
	{
	char *s;
	write(1,msg,strlen(msg));
	s = line;
	while(read(0,s,1) && *s != '\r') {
		if (*s == '\b' && s > line) {
			write(1,"\b \b",3);
			*s-- = '\0';
			}
		else if (*s != '\b') {
			write(1,s,1);
			s++;
			}
		}
	*s = '\0';
	return(0);
	}

/* extract the basic program */

int
get(line,fd)
char *line;		/* contains file name */
int fd;			/* fd to read from */
	{
	char c;
	FILE *out;
	int start=0, end=99999;
	void (*old)();

	sscanf(line,"%s %d %d",buff,&start,&end);
	out = fopen(buff,"w");

	if (!out) {
		fprintf(stderr,"Can't open %s\n",buff);
		return(-1);
		}

	old = signal(SIGALRM,catch);
	siginterrupt(SIGALRM,1);

	fprintf(stderr,"\r\nSaving to %s\r\n",buff);
	if (start > 0)
		sprintf(buff,"list %d,%d\n",start,end);
	else
		sprintf(buff,"list\n");
	put_line(buff,fd,0);

	timeout=0;
	while(!timeout) {
		alarm(2);			/* max time to read character line */
		read(fd,&c,1);	
		c &= 0x7f;
		if (c=='\r') {
			putc('.',stderr);
			alarm(2);
			}
		else
			putc(c,out);
		}
	putc('\n',out);
	fclose(out);
	fprintf(stderr,"\r\ndone\r\n");
	signal(SIGALRM,old);
	return(0);
	}

char lit[6];

int
write_literal(fd,c)
int fd;		/* where to write it */
char c;		/* what to write */
	{
	char *p = lit;
	int nl=0;

	if (c&0x80) {
		*p++ = 'M';
		*p++ = '-';
		c &= 0x7F;
		}
	if (c=='\n')
		nl++;
	if (c < 0x20) {
		*p++ = '^';
		c |= 0x40;
		}

	*p++ = c;
	if (nl) {
		*p++ = '\r';
		*p++ = '\n';
		}
	write(fd,lit,p-lit);
	return(1);
	}

/* catch the alarm timer */

static int
catch()
	{
	timeout = 1;
	dprintf(stderr,"  phone_subs: timeout\n");
	return(0);
	}
