/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* routines using dsp/ dialing stuff (S. A. Uhler) */

#include <stdio.h>
#include "audio_lib.h"
#include "progress.h"

#define dprintf      if (debug) fprintf
#define Dprintf      if (debug>1) fprintf
#define qprintf(x,y)		((int)monitor?(*monitor)(x,y):1)

#define MIN			(HUNK/20)	/* # of loud samples in quiet passage */
#define QUIET_COUNT	3

static int debug = 0;
static char buff[HUNK];		/* audio buffer */
int quiet_thres = ANS_THRES;	/* make extern so apps can futz with it */
int quiet_count = QUIET_COUNT;	/* number of non quiet hunks */

/* wait for dial tone */

int
waitfor_dial(audio_fd,tm)
int audio_fd;			/* fd to get audio from */
int tm;			/* max time to wait 10ths of seconds */
	{
	int mask,n;
	int result = 0;

	progress(buff,0,0);		/* initialize */
	dprintf(stderr,"Waiting for dial tone (up to %d/10 sec)\n",tm);
	set_timeout(tm);
	record_start(audio_fd);
	while(!is_timeout()) {
		mask = DIAL_TONE;
		n = audio_read(audio_fd,buff,sizeof(buff));
		progress(buff,n,&mask);
		if (mask == DIAL_TONE) {
			dprintf(stderr,"s: got dial tone\n");
			result = 1;
			break;
			}
		}
	record_stop(audio_fd);
	audio_flush(audio_fd,0);
	set_timeout(0);
	return(result);
	}

/* dial a number
 * 0-9: tones
 * #    Wait for auxilliary dial tone
 * -    1/2 second pause
 */

int
dial(audio_fd,number)
int audio_fd;
char *number;		/* number to dial */
	{
	unsigned char *tone();
	static int slow;

	slow = getenv("SLOW_DIAL");
	dprintf(stderr,"c: dialing %s %s\n",number,slow?"slow":"fast");
	play_start(audio_fd);
	while(*number) {
		if (*number == '-') {		/* pause for 1/2 second */
			dprintf(stderr,"Pausing (in dial)\n");
			audio_pause(audio_fd,5);
			}
		else if (*number == '#') {		/* wait for dial tone */
			audio_drain(audio_fd);
			dprintf(stderr,"waiting for aux. dial tone\n");
			if (!waitfor_dial(audio_fd,50)) {
				dprintf(stderr,"c: No secondary dial-tone, bye\n");
				return(0);
				}
			audio_pause(audio_fd,1);
			}
		else {
			audio_tone(audio_fd,*number);
			audio_pause(audio_fd,slow?3:1);							/* 100 ms */
			/* audio_drain(audio_fd); */
			dprintf(stderr," %c\n",*number);
			}
		number++;
		}
	dprintf(stderr,"Done Dialing - draining\n");
	audio_drain(audio_fd);

	/* check for dial tone (just in case)

	dprintf(stderr,"Checking for dial tone (just in case)\n");
	record_start(audio_fd);
	n = audio_read(audio_fd,buff,HUNK);
	record_stop(audio_fd);
	if (is_dial(buff,n)) {
		dprintf(stderr,"Still have dial tone after dialing\n");
		return(0);
		}
*/
	return(1);
	}

/* wait for answer after dialing */

/* result:
 *  -2: Busy signal
 *  -1: never heard ringing or ringing stopped before max # of rings 
 *   0: No answer after max # of rings
 *   1: Picked up
 */

#define TIMEOUT		90				/* how long to wait for 1st ring */

int
waitfor_answer(audio_fd,rings,monitor)
int audio_fd;
int rings;								/* how long to wait */
int (*monitor)();						/* call-back function */
	{
	int default_monitor();
	int done = 0;
	int ring = 0;						/* # of rings do far */
	int ringing = 0;					/* true if ringing */
	int quiet=0;						/* silence count */
	int n;								/* bytes read from audio */
	int result = 0;

	if ((int)monitor < 32)	/* backward compatibility */
		monitor=default_monitor;

	dprintf(stderr,"In waitfor_answer\n");

	set_timeout(TIMEOUT);				/* wait 8 seconds for 1st ring */
	record_start(audio_fd);
	while(!done) {
		int mask = ringing ? NONE : RINGING|BUSY;		/* ringing or busy */
		n = audio_read(audio_fd,buff,HUNK);

		if (is_timeout()) {
			qprintf("r: Ringing stopped",ring);
			result = -1;
			break;
			}
		progress(buff,n,&mask);
		switch(mask) {
			case RINGING:
				dprintf(stderr,"Ringing\n");
				if (!ringing) {
					set_timeout(0);
					ring++;
					ringing=1;
					if (!qprintf("R: ringing",ring)) {
						result = -3; break;
						}
					if (ring > rings) {
						qprintf("N: no answer", ring);
						dprintf(stderr,"\nNo answer after %d rings\n",ring);
						result=0;
						done++;
						}
					}
				break;
			case BUSY:
				qprintf("b: busy",ring);
				result=-2;
				done++;
				break;
			case NONE:		/* no tone */
				if (ringing) {
					ringing = quiet=0;
					dprintf(stderr,"Ringing OFF %d/%d\n",ring,rings);
					set_timeout(TIMEOUT);
					}
				break;
			default:
				if (!ringing  && !is_quiet(buff,n,quiet_thres,MIN) &&
									quiet++ > quiet_count) {
					dprintf(stderr,"Picked up\n");
					result=1;
					done++;
					}
				else
					Dprintf(stderr,"quiet level %d\n",quiet);
				break;
			}
		}
	set_timeout(0);	
	record_stop(audio_fd);
	audio_flush(audio_fd,0);
	return(result);
	}

/* return true if all audio samples are quiet */

int
is_quiet(line,n,thres,min)
register unsigned char *line;	/* samples */
int n;				/* number of samples */
int thres;			/* softest sample considered "loud" */
int min;				/* max # of loud samples to accept for "soft" passage */
	{
	register unsigned char *end = line + n;
	register unsigned char t = thres;			/* should be f(thres) */
	register unsigned char c;						/* current sample */
	register int loud = 500;						/* loudest sample so far */
	int count = 0;										/* # of loud samples */

	while(line < end) {
		c = 0x7f & *line++;
		if (c < loud)
			loud = c;
		/* if (c<t) fprintf(stderr,"<%d> ",count); */
		if (c < t && count++ > min) {
			Dprintf(stderr,"A: (%d,%d)%c",loud,line-end+n,' ');
			dprintf(stderr,"a: !quiet\n");
			return(0);
			}
		}
	Dprintf(stderr,"c: quiet\n");
	return(1);
	}

void
debug_dsp(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"dsp_debug is %d\n",n);
	}

int
default_monitor(code, other)
char *code;		/* reason for progress */
int other;		/* supplementary info */
	{
	fprintf(stderr,"%s (ring %d)\n",code+3,other);
	return(1);
	}
