/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/*
**	simpleoratorpipe.c (modified by SAU (from namon)
**
**	Copyright Bellcore 1991. All rights reserved.
**
**	ORATOR is a trademark of Bellcore.
*/

#include	<stdio.h>
#include <sys/signal.h>
#include	<sys/file.h>
#include	<sys/fcntl.h>

#define dprintf      if (debug) fprintf
#define Dprintf      if (debug>1) fprintf
#ifndef DEMI
#define DEMI		"/usr/local/orator"		/* defauult orator path */
#endif

#define	ORATOR_PIPE_FILE		"/usr/tmp/OratorSpeech"
#define	ORATOR_SPEECH_FILE	"/usr/tmp/speech"
#ifndef PROD
#define	ORATOR_TEXT_CMD		 "cd /tmp; textnormal | spellit | namepro -h | syll -b | linguist -s %d | lpced -s | synth -m -a 80 -P %s"
#else
#define	ORATOR_TEXT_CMD		 "textnormal | spellit | namepro -h | syll -b | linguist -s %d | lpced -s | synth -m -a 80 -P %s"
#endif

/* orator cookie, gets passed around as a char * */

struct or {
	FILE *process;					/* popen'd file pointer */
	int  read_fd;					/* fd to read */
	int  lock_fd;					/* fd to write */
	int count;						/* number of synth'd but unread messages */
	int total;						/* total # of synths */
	char fifo[512];				/* path of fifo */
	char lock[512];				/* path of lock file */
	};

char *or_open();
char *getenv();
char buff[1000];

static int debug = 0;

/* test program */

#ifdef MAIN

char line [256];
int catch();
char *s;

main()
	{
	int fd;			/* orator fd for reading voice */
	int n;

	signal(SIGINT,catch);
	signal(SIGTERM,catch);
	signal(SIGHUP,catch);

	s = or_open(NULL,0);
	while(gets(line))  {
		if (strlen(line)<3)
			continue;
		strcat(line,".");		/* just in case */
		or_synth(s,line);
		fd = or_getfd(s);
		while((n=read(fd,buff,sizeof(buff)))>0) {
#ifdef NO_MU
			n=o2mu(buff,n);
#endif
			write(1,buff,n);
			}
		or_end(s);
		}
	or_close(s);
	}

int catch(n)
	{
	or_close(s);
	exit(n);
	}
	
#endif

/* open an orator process, return a magic cookie */

char *
or_open(dir,speed)
char *dir;							/* demi-dir directory */
int speed;							/* synth speed */
{
	char	cmdstr[1024];
	struct or *or = (struct or *) malloc(sizeof (struct or));

	if (speed <=0)
		speed = 80;

	if (dir || !getenv("DEMI_DIR")) {
		sprintf(cmdstr,"DEMI_DIR=%s",dir?dir:DEMI);
		putenv(cmdstr);
		}
	dprintf(stderr,"o: Using DEMI_DIR=%s\n",getenv("DEMI_DIR"));

	putenv("NO_NOTICE=on");
	
	sprintf(or->fifo,"%s.%d",ORATOR_PIPE_FILE,getpid());
	if (mkfifo(or->fifo, 0666)) {
		perror(or->fifo);
		free(or);
		return(0);
		}
   sprintf(or->lock,"%s.%d.lock",ORATOR_PIPE_FILE,getpid());
   if (mkfifo(or->lock, 0666)) {
		perror(or->lock);
		free(or);
		return(0);
		}

	or->lock_fd = 0;
	or->read_fd = 0;
	or->count = 0;
	or->total = 0;
	
	dprintf(stderr,"o: starting Orator\n");
	sprintf(cmdstr, ORATOR_TEXT_CMD, speed, or->fifo);
	if (or->process=popen(cmdstr, "w"))
		return((char *)or);
	else {
		free(or);
		return(NULL);
		}
	}

/* shut down an orator process */

int

or_close(s)
char *s;
	{
	struct or *or = (struct or *) s;

	if (!or)
		return(-1);
	dprintf(stderr,"o: Closing Orator..."); if (debug) fflush(stderr);
	if (or->lock_fd) close(or->lock_fd);
	if (or->read_fd) close(or->read_fd);
	pclose(or->process);
	unlink(or->fifo);
	unlink(or->lock);
	free(or);
	dprintf(stderr,"  done\n");
	return(0);
	}

/* synthesize a phrase */

int
or_synth(s,text)
char *s;
char *text;			/* text to synthesize */
	{
	struct or *or = (struct or *) s;
	fputs(text, or->process);
	fputs(".\n\04\n", or->process);	/* Signals End of Utterance */
	fflush(or->process);
	or->count++;
	or->total++;
	dprintf(stderr,"o %d:  Synthesizing: %s (%d pending)\n",
				or->total,text,or->count);
	return(1);
	}

/* return fd to read synth'd speech from */

int
or_getfd(s)
char *s;
	{
	struct or *or = (struct or *) s;

	dprintf(stderr,"o: getting fd to play speech..."); if (debug) fflush(stderr);
	if (or->lock_fd || or->read_fd) {
		dprintf(stderr,"o: Or_get: can't open or fd, already opened\n");
		return(0);
		}
	
	if (or->count <=0) {
		dprintf(stderr,"o: Or_get: Nothing synth'd\n");
		return(0);
		}

	if ((or->lock_fd = open(or->lock, O_RDONLY|O_NDELAY)) < 0) {
		perror(or->lock);
		return(0);
		}

	or->read_fd = open(or->fifo, O_RDONLY);
	flock(or->lock_fd, LOCK_EX);
	return(or->read_fd);
	}

/* close speech file after reading it all */

int
or_end(s)
char *s;
	{
	struct or *or = (struct or *) s;
	int status = (or->lock_fd>0 && or->read_fd>0);

	dprintf(stderr,"o: utterance done %s\n",status ? "OK":"FAILED");
	if (or->lock_fd > 0) close(or->lock_fd);
	if (or->read_fd > 0) close(or->read_fd);
	or->read_fd = or->lock_fd = 0;
	if (status && or->count>0)
		or->count--;
	return(status);
	}

/* flush an orator message */

int
or_flush(s)
char *s;
	{
	int fd=or_getfd(s);

	if (fd>0) {
		dprintf(stderr,"o: flushing a message\n");
		while(read(fd,buff,sizeof(buff))>0)
			;
		or_end(s);
		}
	return(fd);
	}
		
/* abort any pending orator messages */

int
or_abort(s)
char *s;
	{
	struct or *or = (struct or *) s;
	int result=0;

	dprintf(stderr,"o: Aborting orator\n");
	if (or->read_fd)	{		/* partially read stuff */
		while(read(or->read_fd,buff,sizeof(buff))>0)
			;
		result=1;
		dprintf(stderr,"o: flushing partial message\n");
		or_end(s);
		}

	while(or->count>0) {		/* messages in the Q */
		or_flush(s);
		result++;
		}
	return(result);
	}

/* count of synthed but not yet played messages */

int
or_count(s)
char *s;
	{
	struct or *or = (struct or *) s;
	return(or->count);
	}

void
debug_or(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"orator_debug in %d\n",n);
	}

#ifdef NO_MU
/* convert orator to mu-law (whimpy version) use multiples of 10 bytes */

extern unsigned char *Sb2i;

int
o2mu(buff,n)
char *buff;		/* the data */
int n;			/* number of samples */
	{
	char *y = buff;
	short *x = (short *) buff;
	short *end = (short *) (buff+n);
	int i,cnt=0;
	

	/* convert 5 input samples at a time */

	while(x+5<=end) {
		y[0] =  Sb2i[(4*x[0] + 1*x[1])>>3];
		y[1] =  Sb2i[(3*x[1] + 2*x[2])>>3];
		y[2] =  Sb2i[(2*x[2] + 3*x[3])>>3];
		y[3] =  Sb2i[(1*x[3] + 4*x[4])>>3];
		y+=4;
		x+=5;
		cnt+=4;
		}
#ifdef SOMEDAY
/* extra bytes on the end */
	switch(end-x) {
		case 3:
			break;
		case 2:
			break;
		case 1:
			break;
		}
#endif
	return(cnt);
	}
#endif
