/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* phone interface subs (TAKE II)  sau */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sgtty.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include "phone_lib.h"

/* STIM interface characters - much match up with phone.bas interface  */

#define PH_HANGUP		'd'		/* got disconnect from phone */
#define PH_ALIVE		'A'		/* phone interface is alive */
#define PH_ACK			'Y'		/* command completed OK */
#define PH_NAK			'N'		/* command DID NOT complete OK */
#define PH_ECHO		'X'		/* echo command, for sync. */

#define WAIT			250		/* 100ths-sec to wait for write to port */
#define dprintf		if (debug) fprintf
#define Dprintf		if (debug>1) fprintf
#define dputc			if (debug) putc

#define WHY()			(errno?sys_errlist[errno]:"OK")

extern char *sys_errlist[];
extern int errno;
static int set_mode();

/* open the phone port */

static int timeout = 0;			/* true if we had a timeout */
static int timer = 0;			/* timer is set */
static int alarm_catch();

static char line[80];			/* big enough for anything (yea right!) */
static int debug = 0;
static int do_hup = 0;			/* send self sighup if 'd' returned */
static int (*call)()=NULL;		/* call this instead of sighup if set */
static char *stuff;				/* arg to pass to call */
char *getenv();
char *index();

static 	char *valid = "dDhHeElLpP";		/* valid modes */

static int modem_mode = -1;					/* saved state of modem control */

/* open the phone port */

int
phone_open(port,how,mode)
char *port;
int how;
char *mode;						/* initialize the phone to these modes */
	{
	int off = 0;				/* to enable modem control */
	int fd;

	if (how&PH_SIGHUP)			/* send sighup on DCD down */
		do_hup=getpid();

	if (how&PH_NODELAY) {
		debug_set();					/* set the global debugging characters */
		Dprintf(stderr,"Opening %s NODELAY",port); fflush(stderr);
		fd = open(port,O_RDWR | O_NDELAY);
		Dprintf(stderr," as %d %s\n",fd,WHY());

		/* make sure modem control is enabled */

		if (modem_mode < 0)
			ioctl(fd, TIOCGSOFTCAR, &modem_mode);
		if (modem_mode > off)
			ioctl(fd, TIOCSSOFTCAR, &off);
		}
	else {
		Dprintf(stderr,"Opening %s DELAY ...",port); fflush(stderr);
		fd = open(port,O_RDWR);
		Dprintf(stderr," as %d %s\n",fd,WHY());
		}	

	if (fd<0)  {
		if (debug) perror("phone_open");
		return(-1);
		}

	if (ioctl(fd,TIOCEXCL,0) != 0 && isatty(fd)) {
		dprintf(stderr,"p: open_phone: Can't set %s to exclusive use\n",port);
		close(fd);
		return(-2);
		}
	fcntl(fd,F_SETFD,FD_CLOEXEC);

	set_mode(fd,RAW,ECHO,B9600);

	if (how&PH_NODELAY && how&PH_NOCHECK==0) {		/* force open */
		write(fd,"?",1);
		dprintf(stderr,"p: phone_open() wrote [?] to phone\n");
		phone_read(fd,line,sizeof(line),25);
		if (*line != PH_ALIVE) {
			dprintf(stderr,"p: Phone in unknown state, returned %s\n",line);
			close(fd);
			return(-1);
			}

		/* put phone in known state */

		if (mode && *mode == 'A') {
			int bad = 0;
			for(++mode;*mode;mode++) {
				if (index(valid,*mode) &&
						!index(line,*mode)&&
						!phone_command(fd,*mode)) {
					dprintf(stderr,"p: Phone check can't set %c\n",*mode);
					bad++;
					}
				}
			if (bad)
				return(-bad);
			}
		Dprintf(stderr,"P: PHONE OPENED %s->%d\n",port,fd);
		return(fd);
		}
	return(fd);
	}

/* close the phone port */

int
phone_close(fd)
int fd;
	{
	/* int on=1;		disable modem control */

	if (fd<0)
		return(0);
	
	on_hook(fd);
	dcd_off(fd);

	/* reset modem control modes

	if (modem_mode > 0)
		ioctl(fd, TIOCSSOFTCAR, &modem_mode);

	*/

	sleep(1);		/* just testing */
	close(fd);
	Dprintf(stderr,"P: PHONE CLOSED (%d)\n",fd);
	return(1);
	}

/* set callback function */

char *
set_callback(func,data)
char *data;
int (*func)();
	{
	char *prev = (char *) call;
	call = func;
	stuff=data;
	dprintf(stderr,"p: %s callback function\n",func?"Setting":"Resetting");
	return(prev);
	}

static int
set_mode(file,on,off,speed)
int file;      /* file pointer */
int on;        /* flags to turn on */
int off; 	   /* flags to turn off */
int speed;		/* set baud rate (if not zero) */
{
   struct sgttyb buff;

   if (gtty(file,&buff) != 0)
		perror("getty");
   buff.sg_flags |= on;
   buff.sg_flags &= ~off;
   if (speed)
      buff.sg_ispeed = buff.sg_ospeed = speed;
   if (stty(file,&buff) != 0)
		perror("stty");
        return(0);
	}

/* read until a particular character is recieved, or timeout */

static struct itimerval val = {{0, 0}, {0, 0}};

/* catch the alarm timer */

static int
alarm_catch()
	{
	timeout = 1;
	timer=0;
	Dprintf(stderr,"p: phone timeout\n");
	return(0);
	}

/* read till timeout */

int
Read(fd,data,max,tm)
int fd;     /* file to read from */
char *data; /* data to read */
int max;
int tm;     /* max time to wait (in 10ths seconds) */
   {
   int n;
	static struct itimerval prev;
   void (*old)();

	if (fd <= 0)
		return(-1);

   val.it_value.tv_sec= tm/10;
   val.it_value.tv_usec= (tm%10) * 100000;

	Dprintf(stderr,"P: Reading %d till %d\n",max,tm);
   old = signal(SIGALRM,alarm_catch);
   siginterrupt(SIGALRM,1);
   timeout = 0;
   setitimer(ITIMER_REAL,&val,&prev);
   n = read(fd,data,max);
   alarm(0);
   signal(SIGALRM,old);
   setitimer(ITIMER_REAL,&prev,0);
   return(timeout ? 0 : n);
   }

/* send a single letter command to the phone */

char cmd[3];

int
phone_command(fd,c)
int fd;
char c;
	{
	int n;

	if (fd <= 0)
		return(-1);

	/* tcflush(fd,0);		don't do this */
	write(fd,&c,1);
	dprintf(stderr,"p: sent [%c] to phone\n",c);
	do {
		n=phone_read(fd,cmd,sizeof(cmd),20);
		}
	while (n>=0 && *cmd && index("YN",*cmd) == NULL);
	return(*cmd == PH_ACK);
	}


/* flush the phone fd, return # of events flushed */

int
phone_flush(fd)
int fd;
	{
	int i;

	if (fd <=0)
		return(-1);
	write(fd,"X",1);			/* send sync character */
	dprintf(stderr,"p: syncing phone\n");
	phone_read(fd,cmd,sizeof(cmd),20);
	for (i=0;*cmd!=PH_ECHO;i++)					/* this could block!! */
		phone_read(fd,cmd,sizeof(cmd),20);
	Dprintf(stderr,"P: Flushing done\n");
	return(i);
	}

/* set timer */

int
set_timeout(n)
int n;			/* set a timeout for 10th seconds */
	{
	static int first = 1;

	timeout = 0;
	if (first) {
		first = 0;
		signal(SIGALRM,alarm_catch);
		siginterrupt(SIGALRM,1);
		}

   if (n > 0) {
		ualarm(n*100000,0);
		timer = n;
		Dprintf(stderr,"P: Setting alarm to %d/10 seconds\n",n);
		}
	else
		alarm(0);

	return(1);
	}

/* see if timeout set */

int
is_timer()
	{
	return(timer);
	}

/* see if timeout expired */

int
is_timeout()
	{
	return(timeout==1);
	}

/* read characters up to and including a new-line, strip \r\n */

int
phone_read(fd,data,max,tm)
int fd;		/* phone fd */
char *data;	/* where to put the result */
int max;		/* max # of chars to read */
int tm;		/*  (10th seconds) */
	{
	int count = 0;
	int n;
	int trips = 0;
	int mask;
   struct timeval time;
	char *s;
	
	if (fd <=0)
		return(-1);

	Dprintf(stderr,"p: phone_read: %d chars, %d/10 secs\n",max,tm);
	*data='\0';
	mask = 1<<fd;
   time.tv_sec = tm/10;
   time.tv_usec =(tm%10)*100000;

	errno=0;
	if (select(32,&mask,0,0,tm?&time:NULL)<=0){		/* timeout or error */
		if (errno) {
			dprintf(stderr,"p: phone_read: select failed %s\n",WHY());
			if (errno != EINTR) return(-1);
			}
		return(0);
		}
	
	while(count<max && trips<5) {
		trips++;
		while ((n = read(fd,data+count,max-count)) <= 0) {
			dprintf(stderr,"p: error reading phone %s\n",WHY());
			if (errno != EINTR) break;
			}
		count += n;
		if (data[count-1] == '\n')
			break;
		fsleep();
		}
	if (count>2)
		data[count-2] = '\0';
	else
		data[0] = '\0';
	dprintf(stderr,"p: phone_read got %d (%d tries) [%s]\n",count,trips,data);

	if ((s=index(data,PH_HANGUP)) && *++s < ' ') {
		dprintf(stderr,"Loop disconnect\n");
		if (call) {
			dprintf(stderr,"p: calling callback function\n");
			(*call)(stuff);
			}
		else if (do_hup) {	/* send hangup if we got 'd' */
			dprintf(stderr,"p: Sending hangup signal\n");
			if (kill(do_hup,SIGHUP) !=0)
				perror("kill SIGHUP");
			dprintf(stderr,"p: hangup signal sent\n");
			}
		}
		
	return(count);
	}

/* temporary, non-blocking phone read */

int
phone_read2(fd,data,max)
int fd;		/* phone fd */
char *data;	/* where to put the result */
int max;		/* max # of chars to read */
	{
	int count = 0;
	int n;
	int trips = 0;
	int mask;
   struct timeval time;
	char *s;
	
	if (fd <= 0)
		return(-1);

	Dprintf(stderr,"p: phone_read: %d chars\n",max);
	*data='\0';
	mask = 1<<fd;
   time.tv_sec = 0;
   time.tv_usec = 0;

	errno=0;
	if (select(32,&mask,0,0,&time)<=0){		/* timeout or error */
		if (errno) {
			dprintf(stderr,"p: phone_read: select failed %s\n",WHY());
			if (errno != EINTR) return(-1);
			}
		return(0);
		}
	
	/* old STIM's are slow */

	while(count<max && trips<5) {
		trips++;
		while ((n = read(fd,data+count,max-count)) <= 0) {
			dprintf(stderr,"p: error reading phone %s\n",WHY());
			if (errno != EINTR) break;
			}
		count += n;
		if (data[count-1] == '\n')
			break;
		fsleep();
		}
	if (count>2)
		data[count-2] = '\0';
	else
		data[0] = '\0';
	dprintf(stderr,"p: phone_read got %d (%d tries) [%s]\n",count,trips,data);

	if ((s=index(data,PH_HANGUP)) && *++s < ' ') {
		dprintf(stderr,"Loop disconnect\n");
		if (call) {
			dprintf(stderr,"p: calling callback function\n");
			(*call)(stuff);
			}
		else if (do_hup) {	/* send hangup if we got 'd' */
			dprintf(stderr,"p: Sending hangup signal\n");
			if (kill(do_hup,SIGHUP) !=0)
				perror("kill SIGHUP");
			dprintf(stderr,"p: hangup signal sent\n");
			}
		}
		
	return(count);
	}

/* check the phone, place in desired state */

int
phone_check(name,modes)
char *name;						/* name of the phone device */
char *modes;					/* modes to be in "ledh" */
	{
	int fd;
	char status[20];
	char *mode = modes;
	int n;
	int bad = 0;

	debug_set();					/* set the global debugging characters */

	fd = open(name,O_RDWR | O_NDELAY);
	dprintf(stderr,"p: checking %s for %s\n",name,modes);

	if (fd<0) {
		dprintf(stderr,"p: can't open %s\n",name);
		return(-1);
		}

	set_mode(fd,RAW,ECHO,B9600);
	write(fd,"?",1);
	n = phone_read(fd,status,sizeof(status),25);

	if (n<=0 || *status != PH_ALIVE) {
		dprintf(stderr,"p: cant get %s status\n",name);
		close(fd);
		return(-2);
		}

	dprintf(stderr,"p: setting up phone modes for %s\n",name);
	for(;*mode;mode++) {
		if (index(valid,*mode) &&
				!index(status,*mode)&&
				!phone_command(fd,*mode)) {
			dprintf(stderr,"p: Phone check can't set %c\n",*mode);
			bad++;
			}
		}
	close(fd);
	return(bad?0:1);
	}

int
phone_modes(fd,modes)
int fd;						/* name of the phone fd */
char *modes;					/* modes to be in "ledhp" */
	{
	char status[20];
	int n;
	int bad = 0;

	if (fd <= 0)
		return(-1);
	write(fd,"?",1);
	n = phone_read(fd,status,sizeof(status),25);

	if (n<=0 || *status != PH_ALIVE) {
		dprintf(stderr,"p: cant get status\n");
		return(0);
		}

	dprintf(stderr,"p: setting modes to [%s]\n",modes);
	for(;*modes;modes++) {
		if (index(valid,*modes) &&
				!index(status,*modes)&&
				!phone_command(fd,*modes)) {
			dprintf(stderr,"p: Phone check can't set %c\n",*modes);
			bad++;
			}
		}
	return(bad?-bad:1);
	}

/* do built-in touch tone dialing (no checking for off hook) */

int
_dial(fd,num)
int fd;		/* phone interface */
char *num;	/* number to dial */
	{
	int result=1;

	if (fd <= 0)
		return(-1);

	for(;result && num && *num;num++) {
		if (index("0123456789*#",*num))
			result = phone_command(fd,*num);
		else
			usleep(500000);	/* 1/2 second */
		}
	return(result);
	}

/* turn on debugging */

void
debug_phone(n)
int n;
	{
	if ((debug = n))
		fprintf(stderr,"phone_debug in %d\n",n);
	}
