/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/*
 * generate a fsk tones (hardwired version) for ADSI interface
 * recursive, phase coherent version by SAU. Algorithm by JFK
 * 8/4/92
 * Data format is: <space> <bit0>...<bit7> <mark> ...
 *   mark  is 1200 hz
 *   space is 2200 hz
 *   1200 baud (6-2/3 samples ber bit @ 8000 samples/sec)
 */

#include <stdio.h>

#define OUT(x,s)	*s++ = Sb2i[(int)(x)]		/* convert to mu-law, put result */
#define SCALE	3500.0		/* scale factor (+-1 * SCALE) */
#define MARK	1				/* 1200 hz */
#define SYNC	800			/* # of mark bit samples in front of message */
#define SYNC2	800			/* 100 ms of mark after last byte */

extern unsigned char  *Sb2i;	/* for mu-law conversion */
int counts[] = {7,7,6};			/* table of samples in each bit */

unsigned int marks[] = {0x200, 0x600, 0xE00, 0x1E00, 0x3E00, 0x7E00, 0xFE00}; 

/* magic constants for recursion */

struct konst {
	double a, b, c;
	} ;

static struct konst k[] = {
	0.7159209562, 0.8191014930, -0.3128689301,
	-0.8740320489, 1.2208499296, 1.1755705046,
	};

/* convert text to FSK tones, return # of bytes */

int
to_fsk(text,n,mark_bits,buff)
unsigned char *text;		/* text to convert */
int n;						/* number of bytes */
int mark_bits;				/* number of additional mark bits */
char *buff;					/* where to put the output */
	{
	register char *s = buff;
	register struct konst *kp = k;	/* pointer to constants */
	register double n0, n1, n2;		/* sample values */
	double scale = SCALE;				/* output level */
	int count=SYNC;						/* # of samples in this bit */
	int tone = MARK;						/* current bit value */
	int state = 0;							/* index into samples/bit array */
	int c;									/* "10 bit" character to output */

	if (mark_bits < 0 || mark_bits > 6)
		mark_bits = 0;
	mark_bits = marks[mark_bits];

	/* send count/(6-2/3) bits  of "mark" characters */

	n2 = 0.0;
	n1 = 0.809016994*scale;
	kp = &k[tone];
	while(count-- > 0) {
		n0 = kp->c*n1-n2;
		n2=n1, n1=n0;
		OUT(n0,s);
		}

	/* send the tones for each character space - <8 bits> - mark(s) */

	while(n-- > 0) {
		c = *text++;
		for(c = (c << 1) | mark_bits;c;c>>=1) {
			if (tone != (c&1)) {
				n2 = kp->a*n1 + kp->b*n2;		/* change n2 for new freq */
				tone = 1-tone;
				kp = &k[tone];
				}
			for(count=counts[state];count>0;count--) {
				n0 = kp->c*n1-n2;
				OUT(n0,s);
				n2 = n1, n1 = n0;
				}
			if (state++ == 2)		/* adjust sample count per bit */
				state = 0;
			}
		}

	/* send 100 ms of mark bits */

	for(count=SYNC2;count>0;count--) {
		n0 = kp->c*n1-n2;
		OUT(n0,s);
		n2 = n1, n1 = n0;
		}
	return(s - buff);
	}

#ifdef MAIN

/* test compact fsk stuff */

#include <stdio.h>

char line[255];
char buff[SYNC + SYNC2 + 7*255];

main(argc,argv)
int argc;
char **argv;
   {
   int bits = 0;
	int n;
   if (argc>1)
      bits = atoi(argv[1]);

   while(fgets(line,sizeof(line),stdin)) {
		n = to_fsk(line,strlen(line),bits,buff);
		write(1,buff,n);
		}
   }

#endif
