/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* sample audio control device for 4.1 (S. A. Uhler) */

#include <sys/ioctl.h>
#include <sun/audioio.h>
#include <stdio.h>

#define CNTL		"/dev/audioctl"

struct cmd {
	char *name;		/* command name */
	int arg;			/* # of args */
	};

struct cmd cmd[] = {
   "status", 0,
   "volume", 1,
	"record", 1,
	"monitor", 1,
	"internal", 0,
	"external", 0,
	NULL,0
	};

struct audio_info info, info2;

void
main(argc,argv)
int argc;
char **argv;
	{
	int fd;		/* audio cntl fd */
	int changed = 0;
	register int i;

	if (argc <2) {
		fprintf(stderr,"usage: %s <stuff>\n",*argv);
		for(i=0;cmd[i].name;i++)
			printf("	%s%s\n",cmd[i].name,cmd[i].arg ? " <value>":"");
		exit(1);
		}

	if ((fd = open(CNTL,2)) <0) {
		fprintf(stderr,"Sorry, %s cant open %s\n",*argv,CNTL);
		exit(2);
		}

	AUDIO_INITINFO(&info);

	for(i=1;i<argc;i++) {
		switch(lookup(argv[i])) {
			case 0:		/* status */
				if (changed) {
				   if (ioctl(fd,AUDIO_SETINFO,&info) <0) {
						fprintf(stderr,"Sorry, %s cant set status for %s\n",
									argv[0],CNTL);
						exit(3);
						}
					changed = 0;
					}
				if (ioctl(fd,AUDIO_GETINFO,&info2) <0) {
					fprintf(stderr,"Sorry, %s cant get status for %s\n",*argv,CNTL);
					exit(3);
					}
				printf("record gain %d/%d\n",info2.record.gain,AUDIO_MAX_GAIN);
				printf("playback gain %d/%d\n",info2.play.gain,AUDIO_MAX_GAIN);
				printf("monitor gain %d/%d\n",info2.monitor_gain,AUDIO_MAX_GAIN);
				printf("play port: %s\n",info2.play.port==AUDIO_SPEAKER ?
                   "internal":"external");
				continue;		/* no set */
				break;
			case 1:		/* volume */
				info.play.gain = atoi(argv[++i])&255;
				changed++;
				break;
			case 2:		/* record volume */
				info.record.gain = atoi(argv[++i])&255;
				changed++;
				break;
			case 3:		/* monitor volume */
				info.monitor_gain = atoi(argv[++i])&255;
				changed++;
				break;
			case 4:		/* internal speaker */
				info.play.port = AUDIO_SPEAKER;
				changed++;
				break;
			case 5:		/* external speaker */
				info.play.port = AUDIO_HEADPHONE;
				changed++;
				break;
			default:
				fprintf(stderr,"Invalid command: %s\n",argv[i]);
				continue;
			}
		}

	if (changed && ioctl(fd,AUDIO_SETINFO,&info) <0) {
		fprintf(stderr,"Sorry, %s cant set status for %s\n",argv[0],CNTL);
		exit(3);
		}
	exit(0);
	}

int
lookup(s)
char *s;
	{
	register int i;

	for(i=0;cmd[i].name;i++) {
		if (strcmp(s,cmd[i].name)==0)
			return(i);
		}
	return(-1);
	}
