/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* compress a Sparc audio file (based on toss) S.A. Uhler 11/90 */

#include <stdio.h>

#define dprintf	if (debug)fprintf
#define MAXHUNK		1000			/* max hunk size */
#define HUNK			150			/* Default hunk size */
#define MERGE			20				/* Default merge size */
#define MAXCMP			150			/* Maximum compression % */

#define GET_OPT(i)   \
   strlen(argv[i])>2 ? argv[i]+2 : argv[++i]

unsigned char keep_b[MAXHUNK];
unsigned char merge_b[MAXHUNK];
unsigned char toss_b[MAXHUNK];

int debug=0;

void merge_it();

void
main(argc,argv)
int argc;
char **argv;
	{
	int keep = HUNK;	/* Number of bytes to keep in a hunk */
	int toss;			/* Number of bytes to toss in a hunk */
	int merge = MERGE;	/* Number of bytes to use for linear interpolation */
	int count;			/* # of characters read/written */
   int compress=50;		/* % compression 0 - 150 */
	int do_flush=0;	/* flush after each hunk */
	int i;

	debug = getenv("DEBUG");

	if (argc<2) {
		fprintf(stderr,"usage: ... | %s [-h# -m# -u] %%-compression | ...\n",
				*argv);
		fprintf(stderr," Speed up Sparc audio without changing the pitch\n");
		fprintf(stderr,"  -h#: Hunk size (default is %d)\n", HUNK);
		fprintf(stderr,"  -m#: Bytes to merge between hunks (default is %d)\n",
				MERGE);
		fprintf(stderr,"  -u:  Unbuffered\n");
		exit(1);
		}

	for(i=1;i<argc;i++) {
		if (*argv[i] == '-')
			switch (argv[i][1]) {
				case 'h':				/* get hunk size */
					keep = atoi(GET_OPT(i));
					break;
				case 'm':				/* get merge size */
					merge = atoi(GET_OPT(i));
					break;
				case 'u':				/* unbuffer */
					do_flush++;
					dprintf(stderr,"Turning off buffering\n");
					break;
				default:
					fprintf(stderr,"%s: bad flag %c ignored\n",argv[0],argv[i][1]);
				}
		else
			compress = atoi(argv[i]);
		}

	if (compress<0 || compress > MAXCMP) {
		fprintf(stderr,"Compression out of range: %d - %d\n",0,MAXCMP);
		exit(2);
		}

	toss = keep * compress / 100;
	
	if (merge > keep) {
		fprintf(stderr,"Hunk size must be > than merge size\n");
		exit(3);
		}

	if (merge > toss)
		merge = toss;

	toss -= merge;

	dprintf(stderr,"Compression %d%%: keep: %d toss: %d merge: %d\n",
				compress,keep,toss,merge);
	dprintf(stderr,"Actual compression: %d\n",(toss+merge)*100/keep);

	while(!feof(stdin)) {
		count = fread(keep_b,1,keep,stdin);
		fwrite(keep_b,1,keep-merge,stdout);
		if (count<0)
			break;

		count=fread(toss_b,1,toss,stdin);
		if (count<0)
			break;

		if (merge > 0) {
			count = fread(merge_b,1,merge,stdin);
			merge_it(merge_b,keep_b+keep-merge,merge);
			fwrite(merge_b,1,merge,stdout);
			if (count<merge) break;
			}
		if (do_flush)
			fflush(stdout);
		}
	exit(0);
	}

/* merge 2 sparc audio data streams using linear interpolation */

extern unsigned char  Ub2i[];            /* unsigned 13 bit binary to mu-law */
extern unsigned short I2ub[];             /* mu-law to unsigned 13 bit binary */

void
merge_it(dst,src,count)
register unsigned char *dst;		/* destination buffer */
register unsigned char *src;		/* src buffer */
int count;					/* number of bytes to merge */
	{
	register int i,a,b;
	unsigned short c;

	for(i=0;i<count;i++,src++,dst++) {
		a = i+1;
		b = count+1-a;
		c = (I2ub[*dst]*a + I2ub[*src]*b)/(a+b);
		*dst = Ub2i[c];
		}
	}
