/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* routines for ADSI data mode (S A Uhler 11/2/92)
 * this should probably be redone as a state machine
 * The protocol has been hacked to work with the panasonic ADSI (like)
 * telephone
 */

#include <stdio.h>
#include <sys/file.h>		/* temporary */

#define TYPE_ADSI		0x84		/* ADSI message */
#define MSG_MAX		256		/* maximum total length */
#define SYNC			900		/* size of leading/trailing sync bits */
#define MAGIC			"Mb"		/* arbitrary random adsi msg buff id */
#define TONE_SAMPLES	1000		/* number of samples of touch tone data */
#define TONE_VOLUME	75			/* tone volume in % */
#define ACK_TIMEOUT	5			/* time to wait for DTMF acknowlege */

static unsigned char buff[2*SYNC + 7*MSG_MAX];	/* place to hold audio equiv */
int no_ack = 0;					/* set to 1 for testing */

static void print(unsigned char *s,int n); /* !!proto */
static unsigned char cksum(unsigned char *s,int n);	/* !!proto */ 
int free_msg(unsigned char *s); /* !!proto */ 
int send_msg(unsigned char *msg,int len,int audio_fd,int phone_fd);/* !!proto */
unsigned char *get_msgbuf(); /* !!proto */ 

/* send an adsi message, with proper handshaking */

int
send_msg(msg,len,audio_fd,phone_fd)
unsigned char *msg;		/* message to send (need 2 "free" bytes before message) */
int len;						/*	message length */
int audio_fd;				/* audio file descriptor */
int phone_fd;				/* phone file descriptor */
	{
	int n;
	static int count = 0;
	char path[50];
	int save;

	sprintf(path,"/tmp/save.%d.au",count++);
	save = open(path,O_WRONLY|O_CREAT,0600);

	if (len <0 || len > 254) {
		dprintf("a","Invalid length\n");
		return(-1);
		}

	/* encapsulate message */

	msg[-2] = TYPE_ADSI;
	msg[-1] = len;
	msg[len] = cksum(msg-2,len+2);

	/* send the message - set to data mode */

	n = phone_flush(phone_fd);
	audio_drain(audio_fd);
	dprintf("a","=> data mode(flushed %d tones)\n",n);
	gen_tone('A',TONE_SAMPLES,TONE_VOLUME,buff);
	audio_write(audio_fd,buff,TONE_SAMPLES);
	write(save,buff,TONE_SAMPLES);
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	audio_tone(save,'-');
	n = phone_read(phone_fd,buff,3,ACK_TIMEOUT);
	if (n==0 || *buff != 'A') {
		dprintf("a","DTMF detection is not enabled!\n");
		if (!no_ack) return(-2);
		}

	/* get data mode ACK */

	n = phone_read(phone_fd,buff,3,ACK_TIMEOUT);
	if (n==0 || *buff != 'D') {
		dprintf("a","Switch to data mode not acknowledged (D != %c)\n",
			n>0?*buff:'T');
		phone_flush(phone_fd);
		if (!no_ack) return(0);
		}
	n = phone_read(phone_fd,buff,3,ACK_TIMEOUT);		/* get the '2' */
	dprintf("a","In data mode (D2 == D%c)\n",n>0?*buff:'T');
	phone_flush(phone_fd);	/* make sure tone is finished */

	/* send FSK message */

	print(msg-2,len+3);
   n = to_fsk(msg-2,len+3,0,buff); 	/* generate FSK tones */
	dprintf("a","sending %d bytes of audio\n",n);
	audio_write(audio_fd,buff,n);
	write(save,buff,n);
	audio_tone(save,'-');
	audio_tone(save,'-');
	close(save);
	
	/* get message ACK */

	n = phone_read(phone_fd,buff,3,ACK_TIMEOUT);
	if (n==0 || *buff != 'D') {
		dprintf("a","Switch to data mode not acknowledged (D != %c)\n",
			n>0?*buff:'T');
		phone_flush(phone_fd);
		if (!no_ack) return(0);
		}

	n = phone_read(phone_fd,buff,3,ACK_TIMEOUT);
	if (n==0 || *buff != '1') {
		dprintf("a","Message corrupted (D1=D%c)\n", n>0?*buff:'T');
		phone_flush(phone_fd);
		if (!no_ack) return(0);
		}
	return(1);
	}

/* get an ADSI message buffer */

unsigned char *
get_msgbuf()
	{
	char *malloc();
	unsigned char *result = (unsigned char *) malloc(2+MSG_MAX);
	
	if (result) {
		bcopy(MAGIC,result,2);
		return(result+4);
		}
	else
		return(NULL);
	}

/* free an adsi message buffer */

int
free_msg(s)
unsigned char *s;
	{
	if (s && bcmp(s-4,MAGIC,2)==0) {
		free(s-4);
		return(1);
		}
	return(0);
	}
	
/* compute ADSI check sum */

static unsigned char
cksum(s,n)
unsigned char *s;		/* message to checksum */
int n;					/* message length */
	{
	register int sum = 0;
	while(n-- > 0)
		sum += *s++;
	return((-sum)&0xFF);
	}

/* print a binary string */

static  void
print(s,n)
unsigned char *s;
int n;
   {
	while (n-- > 0) {
		if (*s < 0x80 && *s >= ' ')
			printf("%c ",*s++);
		else
			printf("%.2x ",*s++);
      }
   putchar('\n');
   }
