/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* get caller ID info from ROCHELLE ANI-232 converter (S A Uhler) */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <termio.h>
#include <stdio.h>

/* open the caller-id port */
	
static struct termio buff;

int
cid_open(port)
char *port;
	{
	int fd;
	int power = TIOCM_LE | TIOCM_DTR | TIOCM_RTS;	/* power is supplied here */

	fd = open(port,O_RDWR | O_NDELAY);

	if (fd<0) {
		perror("open");
		return(-1);
		}

	fcntl(fd,F_SETFD,FD_CLOEXEC);
	buff.c_iflag = IGNBRK|ISTRIP|CLOCAL;
	buff.c_oflag = 0;
	buff.c_lflag = 0;
	buff.c_cflag = B1200|CS8;	/* disable by default */
	buff.c_cc[VMIN] = 60;		/* more than expected */
	buff.c_cc[VTIME] = 2;
   if (ioctl(fd,TCSETA,&buff) != 0) {
		perror("ioctl SET CID MODES");
		close(fd);
		return(-1);
		}

	/* initialization - send a break, wait for "ringing" to pulse on and off */

	if (ioctl(fd,TIOCMSET,&power) != 0)	{ /* power on */
		perror("ioctl POWER ON");
		close(fd);
		return(-1);
		}

	if (!cid_reset(fd)) {
		close(fd);
		return(-1);
		}
	else
		return(fd);
	}

/* reset the interface, do initial handshake (can hang if broken) */

static int timeout = 0;

int
cid_reset(fd)
int fd;
	{
	static void catch();
	int bits;

	void (*old)();

	if (ioctl(fd,TIOCMGET,&bits) != 0) {		/* RI should be off */
		perror("ioctl TIOCMGET - cid_reset");
		return(0);
		}

	if (bits&TIOCM_RNG) {
		fprintf(stderr,"Ringing high at reset????\n");
		return(0);
		}

	if (tcsendbreak(fd, 2)!=0)  {
		perror("send break");
		return(0);
		}

	old = signal(SIGALRM,catch);
	timeout=0;
	alarm(5);
	while(!timeout) {
		if (ioctl(fd,TIOCMGET,&bits) != 0) {
			perror("GET - waiting for ring on");
			return(0);
			}
		fprintf(stderr,"0x%x - 0x%x (on?)\n",bits,TIOCM_CAR);
		if (bits&TIOCM_CAR)		/* should be ringing, but ring -> DCD on
                                 sun's on board serial ports    */
			break;
		}

	while(!timeout) {
		if (ioctl(fd,TIOCMGET,&bits) != 0) {
			perror("GET - cid_reset");
			return(0);
			}
		if (!(bits&TIOCM_CAR))		/* should be ringing, but... */
			break;
		fprintf(stderr,"0x%x - 0x%x (off?)\n",bits,TIOCM_CAR);
		}
	alarm(0);
	signal(SIGALRM,old);

	return(1);
	}

static void
catch()
	{
	timeout++;
	}

/* enable the interface */

int
cid_enable(fd)
int fd;
	{
	int n;

   ioctl(fd,TCGETA,&buff);
	buff.c_cflag |= CREAD;
	n = ioctl(fd,TCFLSH,0);
   n |= ioctl(fd,TCSETA,&buff);
	return(n==0);
	}

/* disable the interface */

int
cid_disable(fd)
int fd;
	{
	int n;

   ioctl(fd,TCGETA,&buff);
	buff.c_cflag &= ~CREAD;
   n = ioctl(fd,TCSETA,&buff);
	return(n==0);
	}

/* get the caller ID string */

char * 
cid_read(fd,line)
int fd;
char *line;
	{
   struct timeval time;
	int mask = 1<<fd;
	int i;
	int n;

	time.tv_sec = 4;
	time.tv_usec = 0;

	cid_enable(fd);

	bzero(line,60);
	if (select(32,&mask,0,0,&time) && mask && 
				(n = read(fd,line,60)) > 10) {
		for(i=0;i<n;i++)
			if (isdigit(line[i]))
				break;
		i +=8;
		if (i<n && isdigit(line[i])) {		/* got a phone number */
			line[i+10] = '\0';
			return(line+i);
			}
		}
	return(NULL);
	}

#ifdef MAIN

char line[100];

main(argc,argv)
int argc;
char **argv;
	{
	int fd = cid_open(argc>1?argv[1]:"/dev/ttya");
	printf("waiting up to 4 seconds for cid info\n");
	while (1) {
		cid_read(fd,line);
		if (line) printf("got: %s\n",line);
		}
	}
#endif
