/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* sample TCL intepreter for Phone applications  - commands (S. A. Uhler) */

#include <stdio.h>
#include <sys/signal.h>
#include <sys/file.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include "tcl.h"
#include "defs.h"
#include "phone_lib.h"

struct commands {
	char *name;						/* name of command */
	Tcl_CmdProc *func;	/* name of function to call */
	ClientData data;				/* data to pass to function */
	};

/* commands */

int adsi(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int audio(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int call(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int cnv2tt(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int db(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int debug(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int do_sleep(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int on(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int phone(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int play(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int prdate(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int record(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int synth(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int t_echo(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
int t_hangup(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
#ifndef NOGDBM
int gdbm(ClientData cd, Tcl_Interp *it, int argc, char *argv[]); /* !!proto */
#endif

static int copy_file(int from,int to);	/* !!proto */
char *or_open(char *dir, int speed);

struct commands commands[] = {
	"adsi",adsi, (ClientData) &stuff_data,		/* adsi interface stuff */
	"audio",audio, (ClientData) &stuff_data,	/* audio control stuff */
	"beep",audio, (ClientData) &stuff_data,	/* audio control stuff */
	"call",call,(ClientData) &stuff_data,		/* place a call */
	"cnv2tt",cnv2tt, NULL,							/* convert string to touch tones */
	"db",db,(ClientData) &stuff_data,			/* database operations */
#ifndef NOGDBM
	"dbm",gdbm,(ClientData) &stuff_data,		/* experimental hashed db */
#endif
	"debug",debug,(ClientData) &stuff_data,	/* run interactively for a while */
	"echo", t_echo, NULL,							/* echo (ala sh) */
	"hangup",t_hangup,(ClientData) &stuff_data,	/* go hangup */
	"on",on,(ClientData) &stuff_data,			/* do on commands */
	"phone",phone, (ClientData) &stuff_data,	/* phone control stuff */
	"play",play, (ClientData) &stuff_data,		/* play a file, get tones */
	"prdate",prdate, NULL,							/* print the time/date */
	"record",record, (ClientData) &stuff_data,	/* record a message */
	"sleep",do_sleep, NULL,							   /* sleep some */
	"synth",synth,(ClientData) &stuff_data,	/* orator synthesis */
	NULL, NULL, NULL
	};

/* add PhoneScript commands to the TCL interpreter */

void
init_commands(interp)
Tcl_Interp *interp;
	{
	struct commands *cmds = commands;

	for(;cmds->name;cmds++)
		Tcl_CreateCommand(interp,cmds->name,cmds->func,cmds->data,NULL);
	}

/* echo (ala sh) */

static FILE *last_file;					/* last file opened for echo */
static char *last_name=NULL;			/* name of last file used */

int
t_echo(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
   int i;
	char *nonl;
	char *echo_file;
	FILE *file;

	/* dprintf("T","In echo [%s...]\n",argv[1]); */
	argc = parse_args(argv);
	nonl=get_value(interp,"nonl");
	echo_file=get_value(interp,"echo_file");

	/* where to output to */

	if (strcmp(echo_file,"stderr")==0)
		file = stderr;
	else if (*echo_file) {
		dprintf("e","Echo outputting to %s\n",echo_file);
		if (!last_name || strcmp(last_name,echo_file)!=0) {
			dprintf("e","Echo caching file named %s\n",echo_file);
			if (last_name) {
				fclose(last_file);
				dprintf("e","Echo freeing caching file named %s\n",last_name);
				free(last_name);
				last_name = NULL;
				}
			if ((last_file = fopen(echo_file,"w")) == NULL) {
				Tcl_AppendResult(interp,*argv,
							": can't open file ",echo_file,"\n",NULL);
				return(TCL_ERROR);
				}
			last_name = save_line(echo_file);
			}
		file = last_file;
		}
	else
		file = stdout;

	for(i=1;argv[i];i++) {
		if (i>1) fprintf(file," ");
		fputs(argv[i], file);
		}
   if (is_true(nonl)!=1)
		fprintf(file,"\n");
	fflush(file);

   return(TCL_OK);
   }

/* names of on commands */

static char *on_names[] = {
	"call",			/* a call is answered */
	"endringing",	/* ringing stopped before call was answered */
	"hangup",		/* call was completed */
	"int",			/* a ^C happened */
	"ringing",		/* ringing started */
	"signal",		/* got a signal */
	"error",			/* got a TCL error */
	"extension",	/* got an extension pickup */
	NULL
	};

/* process "on" conditions */

int
on(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	int i;

	argc = parse_args(argv);
	if (argc <2 || argc > 3) {
		Tcl_AppendResult(interp,*argv,": Wrong number of arguments\n",NULL);
		return(TCL_ERROR);
		}
	dprintf("T","In on as %s\n",argv[1],argc);

	for(i=0;on_names[i];i++)
		if (strcmp(argv[1],on_names[i])==0)
			break;

	if (!on_names[i]) {
		Tcl_AppendResult(interp,*argv,": invalid on condition: ",
					argv[i],"\n",NULL);
		return(TCL_ERROR);
		}

	if (argv[2])
		Tcl_SetVar2(interp,"on",argv[1],argv[2],0); 
	else
		Tcl_UnsetVar2(interp,"on",argv[1],0);
	return(TCL_OK);
	}

/* the hangup command */
		
int
t_hangup(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	dprintf("T","Hanging up\n");
	phone_command(S(phone_fd),'h');
	S(hangup)++;
	return(TCL_UNWIND);
	}
		
/* record a message to a file */

int
record(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	char c;
	struct stuff *stuff = (struct stuff *) clientData;
	int sound_fd;
	int tm;
	int silence;
	int truncate;		/* truncate last 'n' samples */

	argc = parse_args(argv);

	dprintf("T","In record (to %s)\n",argc>1?argv[1]:"???");
	if (argc < 3) {
		Tcl_AppendResult(interp,"Usage: ",*argv,
				 "<file> <tt-list> [record_timeout=n] [silence=n]\n", NULL);
		return(TCL_ERROR);
		}

	if ((sound_fd=open(argv[1],O_WRONLY|O_CREAT,0600)) < 0) {
		Tcl_AppendResult(interp,*argv,": can't open ",argv[1],"\n",NULL);
		return(TCL_ERROR);
		}

	truncate = atoi(get_value(interp,"truncate"));
	tm = atoi(get_value(interp,"record_timeout"));
	if (tm==0)		/* max 10 minutes */
		tm=6000;
	silence = atoi(get_value(interp,"silence"));

	audio_drain(S(audio_fd));	/* wait for any pending audio to clear */
	c = get_file(sound_fd,S(audio_fd),S(phone_fd),argv[2],tm,silence);
	if (c != 'd' && truncate > 0) {
		ftruncate(sound_fd,tell(sound_fd)-truncate);
		dprintf("r","Truncating recorded file by %d bytes\n",truncate);
		}
	close(sound_fd);
	Tcl_SetResult(interp,sprintf(line,"%c",c),TCL_VOLATILE);
	if (S(hangup))
		return(TCL_UNWIND);
	else
		return(TCL_OK);
	}

/* sleep */

int
do_sleep(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	double atof();
	struct timeval wait;
	int time;

	if (argc != 2) {
		Tcl_AppendResult(interp,"Usage:",*argv," <seconds>\n",NULL);
		return(TCL_ERROR);
		}

	time = 100.0 * atof(argv[1]) + 0.005;
	if (time > 0) {
		wait.tv_sec = time/100;
		wait.tv_usec = (time%100) * 10000;
   	select(0,0,0,0,&wait);
		}
	return(TCL_OK);
	}

/* print the date */

int
prdate(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	time_t now;
	char *fmt;
	
	dprintf("T","In prdate\n");
	switch(argc) {
		case 1:				/* fetch the current time (in seconds) */
			now = time(NULL);
			Tcl_SetResult(interp,sprintf(line,"%d",now),TCL_VOLATILE);
			return(TCL_OK);
		case 2:
			now = time(NULL);
			fmt = argv[1];
			break;
		case 3: 
			now = (time_t) atoi(argv[1]);	
			fmt = argv[2];
			break;
		default:
			Tcl_AppendResult(interp,"Usage:",*argv,"[time] <format>\n",NULL);
			return(TCL_ERROR);
		}
	strftime(line,TCL_RESULT_SIZE,fmt,localtime(&now));
	Tcl_SetResult(interp,line,TCL_VOLATILE);
	return(TCL_OK);
	}

/* audio management (beep) */

int
audio(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	struct stuff *stuff = (struct stuff *) clientData;
	
	dprintf("T","In audio\n");
	if (argc == 1 && strcmp("beep",argv[0]) == 0) {
		audio_beep(S(audio_fd),2,0);
		return(TCL_OK);
		}
	if (argc>1 && strcmp("audio",argv[0])==0) {
		if (argc>2 && strcmp(argv[1],"play")==0) {	/* set play volume */
			audio_set(S(audio_fd),0,atoi(argv[2]));
			return(TCL_OK);
			}
		if (argc>2 && strcmp(argv[1],"record")==0) {	/* set record volume */
			audio_set(S(audio_fd),atoi(argv[2]),0);
			return(TCL_OK);
			}
		if (strcmp(argv[1],"drain")==0) {	/* drain audio */
			audio_drain(S(audio_fd));
			return(TCL_OK);
			}
		if (strcmp(argv[1],"close")==0) {	/* close audio */
			if (S(audio_fd)>0) close(S(audio_fd));
			S(audio_fd) = -1;
			return(TCL_OK);
			}
		if (strcmp(argv[1],"open")==0) {		/* open audio */
			close(S(audio_fd));
			if (S(audio_fd) == -1)
				S(audio_fd) = audio_open(2);
			return(TCL_OK);
			}
		Tcl_AppendResult(interp,sprintf(line,"Usage: %s",*argv),NULL);
		return(TCL_ERROR);
		}
	Tcl_AppendResult(interp,sprintf(line,"Usage: %s",*argv),NULL);
	return(TCL_ERROR);
	}

/* do database stuff */

#define NEED_DB	1	/* need to have a database selected */

struct do_db {
	char *option;		/* option to perform */
	char min;				/* minimum # of args */
	char max;				/* max args */
	char flags;			/* flags */
	int (*func)();		/* function to call for option */
	};

int db_set(), sb_close(), db_top(), db_next(), db_fields();
int db_update(), db_insert(), db_delete(), db_select(), db_process();
int db_write(), db_close();

static struct do_db options[] = {
	"close",0,0,1,db_close,	/* close the current database */
	"delete",0,0,1,db_delete,	/* deleter the current database item */
	"fields",0,0,1,db_fields,	/* return list of fields */
	"insert",0,0,1,db_insert,	/* insert a new database item */
	"next",0,0,1,db_next,		/* next item */
	"process",1,1,1,db_process,/* process all rows in a db */
	"select",1,2,1,db_select,	/* select tuples */
	"set",1,1,0,db_set,		/* set the current database (open if not opened) */
	"top",0,0,1,db_top, 		/* go to top */
	"update",0,0,1,db_update, 	/* update the current tuple in the database */
	"write",0,1,1,db_write,/* write the current database (open if not opened) */
	NULL,
	};

/* do a database function */

int
db(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	struct stuff *stuff = (struct stuff *) clientData;
	struct do_db *option = options;
	int i;

	argc = parse_args(argv);

	dprintf("T","In db %s\n",argv[1]);
	if (argc < 2) {
		Tcl_AppendResult(interp,"Usage: db <option> [parameters...]\n",NULL);
		return(TCL_ERROR);
		}

	dprintf("d","Running database command with %d args:\n",argc);
	for (i=0;i<argc;i++)
		dprintf("D","  %d:[%s]\n",i,argv[i]);
	
	for(;option->option;option++)
		if (strcmp(argv[1],option->option)==0) {
			if (argc-2 < option->min || argc-2 > option->max) {
				dprintf("d","%s needs %d-%d args, got %d\n",
					argv[1], option->min,option->max,argc-2);
				Tcl_AppendResult(interp,"Wrong number of arguments for DB option\n",NULL);
				return(TCL_ERROR);
				}
			if (!S(db) && option->flags) {
				Tcl_AppendResult(interp,"db ",argv[1],": no database set\n",NULL);
				return(TCL_ERROR);
				}
			return((*option->func)(interp,stuff,argv));
			}
	
	Tcl_AppendResult(interp,"db ",argv[1],": invalid option\n",NULL);
	return(TCL_ERROR);
	}

/* convert an arbitrary character string into touch tones */

char cnvt_tt[] = {
	2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,0,7,7,8,8,8,9,9,9,0
	};

int
cnv2tt(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	char *out;
	int n;
	char c;
	char *s = argv[1];

	argc = parse_args(argv);

	dprintf("T","In cnv2tt\n");
	if (argc != 2) {
		sprintf(line,"Usage: %s <string>\n",*argv);
		Tcl_AppendResult(interp,line,NULL);
		return(TCL_ERROR);
		}
	out = line;
	s = argv[1];

	dprintf("x","Converting [%s] to ",s);
	while(c = *s) {
		if (c>='0' && c<='9')
			*out++ = c;
		else if (c>='A' && c<='Z' && (n=cnvt_tt[*s - 'A']))
				*out++  = '0' + n;
		else if (c>='a' && c<='z' && (n=cnvt_tt[*s - 'a']))
				*out++  = '0' + n;
		else if (c=='#' || c=='*' || c==' ')
			*out++ = c;
		s++;
		}
	*out = '\0';
	Tcl_SetResult(interp,line,TCL_VOLATILE);
	return(TCL_OK);
	}

/* synth an orator string *
 *
 * synth	 [speed=nnn] [flush=n]						initialize orator
 * synth  text												synthesize text
 * synth  text [to] file								synthesize and save
 * synth  # [to] file									render previous synth
 */

int
synth(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	struct stuff *stuff = (struct stuff *) clientData;
	char *getenv();
	static char *temp_dir = NULL;
	int ok;						/* file copy worked */
	int fd;						/* file to contain output */
	int fd_or;					/* file to read synthed. input from */
	int count;
	char *file=NULL;			/* file to dump synth stuff to */

	dprintf("T","In synth [%s]\n",argv[1]);
	argc = parse_args(argv);
	if (!temp_dir) temp_dir = getenv("TMPDIR") ? getenv("TMPDIR") : "/tmp";
	if (!S(or)) {
		int speed = atoi(get_value(interp,"speed"));
   	S(or)=or_open(NULL,speed);
		dprintf(".","Opening orator\n");
      }

	if (!S(or)) {
		Tcl_AppendResult(interp,"Orator  won't start, can't use synth\n",NULL);
		dprintf("s","Opening orator FAILED\n");
		return(TCL_ERROR);
		}

	if (argc == 1) {
		count = or_count(S(or));
		sprintf(line,"%d",count);
		Tcl_SetResult(interp,line,TCL_VOLATILE);
		if (is_true(get_value(interp,"flush")) == 1)
			or_abort(S(or));
		return(TCL_OK);
		}

	if (argc == 2 && strcmp(argv[1],"clear")==0) {	/* clear pending messages */
		or_abort(S(or));
		return(TCL_OK);
		}
	if (argc == 3)
		file = argv[2];
	else if (argc == 4 && !strcmp(argv[2],"to"))
		file = argv[3];
	else if (argc != 2) {
		sprintf(line,"%s: invalid arguments",*argv);
		Tcl_AppendResult(interp,line,NULL);
		return(TCL_ERROR);
		}

	if (strcmp(argv[1],"#")!=0) {		/* not a #, synth  */
		or_synth(S(or),argv[1]);
		}
	else if (or_count(S(or)) <=0) {
		Tcl_SetResult(interp,"0",TCL_VOLATILE);
		return(TCL_OK);
		}

	/* make a temporary name */

	if (file && *file=='%') {
		static int tmp_files = 0;	/* count of temporary files */

		sprintf(line,"%s/PS.%d.%d.au",temp_dir,getpid(),++tmp_files);
		Tcl_SetVar(interp,file+1,line,0);
		file = line;
		Tcl_SetVar(interp,"temp_names",line,
					TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		}

	/* open the file */

	if (file) {								/* dump to file */
		fd = open(file,O_WRONLY|O_CREAT,0600);
		dprintf(".","File %s opened %d\n",file,fd);
		if (fd < 0) {
			sprintf(line,"%s: can't open file %s\n",*argv,file);
			Tcl_AppendResult(interp,line,NULL);
			return(TCL_ERROR);
			}
		if ((fd_or = or_getfd(S(or))) <=0) {
			sprintf(line,"%s: can't read orator file\n",*argv);
			Tcl_AppendResult(interp,line,NULL);
			close(fd);
			return(TCL_ERROR);
			}
		ok = copy_file(fd_or,fd);
		or_end(S(or));
		close(fd);
		if (!ok) {
			sprintf(line,"%s: cant render speech to %s",*argv,file);
			Tcl_AppendResult(interp,line,NULL);
			return(TCL_ERROR);
			}
		}
	Tcl_SetResult(interp,"1",TCL_VOLATILE);
   return(TCL_OK);
   }

/* manipulate the telephone interface directly */

struct phone_options {
	char *name;				/* name of the option */
	char code;					/* command character */
	};

static struct phone_options phone_options[] = {
	"dcd_off", 'd',
	"dcd_on", 'D',
   "blink_fast", 'C',
   "blink_slow", 'c',
	"dtmf_off", 'e',
	"dtmf_on", 'E',
	"flash", 'F',
	"loop_off", 'l',
	"loop_on", 'L',
	"off_hook", 'H',
	"on_hook", 'h',
	"reset", 'R',
	NULL,
	};

int
phone(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	char *index();
	struct stuff *stuff = (struct stuff *) clientData;
	int i;
	int dur;

	dprintf("T","PHONE %s\n",argv[1]);
	argc = parse_args(argv);
	dur = atoi(get_value(interp,"tone_duration"));

	if (!argv[1]) {
		Tcl_AppendResult(interp,"Phone options: ",NULL);
		for(i=0;phone_options[i].name;i++)
			Tcl_AppendResult(interp,"phone_options[i].name",NULL);
		return(TCL_ERROR);
		}

	if (strcmp(argv[1],"open") == 0 && S(phone_fd) == -1) {/* open it */
		if (argv[2] && strcmp(argv[2],"answer") == 0)
			S(phone_fd) = phone_open("/dev/answer",PH_NODELAY,NULL);
		else
			S(phone_fd) = phone_open("/dev/call",PH_NODELAY,NULL);
		return(TCL_OK);
		}

	if (S(phone_fd) == -1) {
		Tcl_AppendResult(interp,"Phone is not open",NULL);
		return(TCL_ERROR);
		}
		
	/* single letter stuff */

	for(i=0;phone_options[i].name;i++) {
		if (strcmp(argv[1],phone_options[i].name)==0) {
			phone_command(S(phone_fd),phone_options[i].code);
			if (S(hangup))
				return(TCL_UNWIND);
			else
				return(TCL_OK);
			}
		}

	/* other stuff */

	if (strcmp(argv[1],"flush") == 0) {				/* flush pending phone input */
		phone_flush(S(phone_fd));
		}
	else if (strcmp(argv[1],"status") == 0) {		/* return phone status */
		write(S(phone_fd),"?",1);
		phone_read(S(phone_fd),line,200,25);
		Tcl_SetResult(interp,line,TCL_VOLATILE);
		}
	else if (strcmp(argv[1],"version") == 0) {	/* return phone version */
		write(S(phone_fd),"V",1);
		phone_read(S(phone_fd),line,200,25);
		Tcl_SetResult(interp,line,TCL_VOLATILE);
		}
	else if (strcmp(argv[1],"tones") == 0 && argv[2]) {	/* play tones */
		play_tones(S(audio_fd),argv[2],argv[3],dur);
		}
	else if (strcmp(argv[1],"dial") == 0 && argv[2]) {	/* dial the phone */
		dial(S(audio_fd),argv[2]);		
		}
	else if (strcmp(argv[1],"close") == 0) {	/* close the phone */
		phone_close(S(phone_fd));
		S(phone_fd) = -1;
		}
	else if (strcmp(argv[1],"answer") == 0 && argv[2]) {	/* answer the phone */
		if (waitfor_answer(S(audio_fd),argv[2],2) == 1)
			Tcl_SetResult(interp,"1",TCL_VOLATILE);
		else
			Tcl_SetResult(interp,"0",TCL_VOLATILE);
      }
	else {													/* invalid command */
		Tcl_AppendResult(interp,"phone: invalid option",NULL);
		return(TCL_ERROR);
		}
	if (S(hangup))
		return(TCL_UNWIND);
	else
		return(TCL_OK);
	}

/* copy a file */

static char buff[BUFSIZ];

static int
copy_file(from,to)
int from;
int to;
	{
	int n;

	while((n=read(from,buff,sizeof(buff)))>0 && write(to,buff,n)==n)
		;
	return(n==0);
	}

/* do the hangup stuff */

int
do_hangup(interp,stuff)
Tcl_Interp *interp;
struct stuff *stuff;
	{
	char *cmd = Tcl_GetVar2(interp,"on","hangup",0);
	char *bad;

	dprintf("T","In do_hangup\n");
	if (cmd) dprintf("S","Running on command hangup\n");
	if (cmd && Tcl_Eval(interp,cmd,0,&bad) == TCL_ERROR) {
		error_cmd(interp,cmd,bad);
		fprintf(stderr,"Error in hangup expression: %s\n",interp->result);
		return(0);
		}	
	if (S(or))
		or_abort(S(or));
	return(1);
	}

/* place a call  we should preserve "hook" state */

static char *call_cmd = NULL;		/* tcl code to run when monitoring calls */
static Tcl_Interp *local_interp;	/* for monitor.  YUK ! */

int
call(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	int dial_tone;				/* time to wait for dial tone */
	int wait_rings;			/* rings tro wait for answer */
	char sharp[2];				/* play a (sharp) tone after answering */
	int result;					/* result of waitfor_answer */
	int on = 0;					/* true if on hook */
	static int monitor();	/* call monitor function */

	dprintf("T","In Call\n");

	*sharp = '\0';
	argc = parse_args(argv);
	dial_tone = atoi(get_value(interp,"dial_tone"));
	wait_rings = atoi(get_value(interp,"wait_rings"));
	*sharp = *get_value(interp,"sharp");

	if (argc<2) {
		Tcl_AppendResult(interp,"usage: ",*argv," <number>\n",NULL);
		return(TCL_ERROR);
		}
	call_cmd = argv[2];
 
	dprintf("c","  Taking the phone off hook\n");
	loop_off(S(phone_fd));
	dtmf_off(S(phone_fd));
	on = off_hook(S(phone_fd));
	if (on) sleep(1);

   /* wait for dial tone */

	dprintf("c","  waiting for dial tone %d/10 sec\n",dial_tone);
   if (!waitfor_dial(S(audio_fd),dial_tone)) {
		Tcl_AppendResult(interp,*argv,": Can't get dial tone\n",NULL);
		if (on) { on_hook(S(phone_fd)); }
		dtmf_on(S(phone_fd));
		loop_on(S(phone_fd));
		return(TCL_ERROR);
      }

	sleep(2);		/* temporary */
   dprintf("c","  got dial tone, dialing %s\n",argv[1]);
   if (!dial(S(audio_fd),argv[1])) {
      dprintf("c","  dialing %s failed\n",argv[1]);
		Tcl_AppendResult(interp,*argv,": Dial tone problems\n",NULL);
		if (on) on_hook(S(phone_fd));
		loop_on(S(phone_fd));
		return(TCL_ERROR);
      }

   dprintf("c","  Waiting %d rings for answer\n",wait_rings);
	local_interp = interp;
   result = waitfor_answer(S(audio_fd),wait_rings, call_cmd?monitor:NULL);
	dtmf_on(S(phone_fd));
	switch(result) {
		case 1:			/* answered OK */
			if (*sharp) {
				dprintf("c","playing \"%c\"\n",*sharp);
				audio_tone(S(audio_fd),*sharp);
				}
			loop_on(S(phone_fd));
			dtmf_on(S(phone_fd));
			return(TCL_OK);
		case 0:			/* no answer */
			Tcl_AppendResult(interp,*argv,": No answer after max rings\n",NULL);
			break;
		case -1:			/* never heard ringing */
			Tcl_AppendResult(interp,*argv,": Never heard ringing\n",NULL);
			break;
		case -2:			/* busy */
			Tcl_AppendResult(interp,*argv,": Got busy signal\n",NULL);
			break;
		}
	if (on) {
		on_hook(S(phone_fd));
		}
	loop_on(S(phone_fd));
	dtmf_on(S(phone_fd));
	return(TCL_ERROR);
	}

/* monitor progress function */

static int
monitor(code,rings)
char *code;		/* reason for call */
int rings;		/* # of rings so far */
	{
	sprintf(line,"%d",rings);
	Tcl_SetVar(local_interp,"reason",code,0);
	Tcl_SetVar(local_interp,"ring",line,0);
	return(Tcl_Eval(local_interp, call_cmd, 0, NULL) == TCL_OK);
	}

/* run PhoneScript interactively (for debugging mode) */

int
debug(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	Tcl_CmdBuf buffer = Tcl_CreateCmdBuf();
	int count = 0;
	char *cmd = "";
	char *bad;
	int result = TCL_OK;

	printf("Entering debug mode, type \"exit\" to resume\n");
	while(!feof(stdin)) {
		if (S(hangup)) {
			do_hangup(interp,stuff);
			S(hangup) = 0;
			}
		if (cmd) count++;
		printf("%d%s ",count,cmd?"%":">"); fflush(stdout);
		fgets(line,MAXLINE,stdin);
		if (strcmp(line,"resume\n")==0) {
			S(resume)++;
			break;
			}
		if (strcmp(line,"exit\n")==0)
			break;
		if ((cmd = Tcl_AssembleCmd(buffer, line)) &&
							 (Tcl_Eval(interp, cmd, 0, &bad) != TCL_OK))  {
			error_cmd(interp,cmd,bad);
			printf("%s\n",interp->result);
			}
		else if (cmd)
			printf("%s\n",interp->result);
		}
	Tcl_DeleteCmdBuf(buffer);
	printf("Leaving debug mode\n");
	return(result);
	}

/* routines for ADSI data mode (sau 11/2/92)
 * usage:
 * adsi clear
 * adsi define id label id label id label ...
 * adsi assign id id ...
 */

/* valid ADSI keywords */

static char *cmds[] = { "clear", "define", "assign", NULL };
static unsigned char clear_msg[] = {0x01, 0x8E, 0x00};
int free_msg(unsigned char *s); /* !!proto */ 
int send_msg(unsigned char *msg,int len,int audio_fd,int phone_fd);/* !!proto */
unsigned char *get_msgbuf(); /* !!proto */ 
static unsigned char msg_buf[256];

/* create an ADSI message
 * This currently only deals with one kind of message.  It should be split
 * in 2 and made more general
 */

int
adsi(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
   {
	struct stuff *stuff = (struct stuff *) clientData;
	unsigned char *msg = msg_buf + 2;
	int i;		/* index */
	int n=0;		/* return code */
	int id;		/* definer ID */
	int len;		/* text length */

   dprintf("T","in ADSI\n");

	argc = parse_args(argv); 

	for(i=0;argc>1 && cmds[i];i++)
		if (strcmp(argv[1],cmds[i])==0)
			break;

	if (argc<2 || !cmds[i]) {
      Tcl_AppendResult(interp,"Usage: ",*argv,
             "clear|define|assign [id label...] [key id...]\n", NULL);
      return(TCL_ERROR);
      }

	/* compose and send the message */

	switch(i){
		case 0:		/* clear */
			bcopy(clear_msg,msg,3);
			n = send_msg(msg,3,S(audio_fd),S(phone_fd));
			break;

		case 1:		/* define */
			if (argc&1 == 1) {		/* args must be in pairs */
				Tcl_AppendResult(interp,
							"adsi define must have even # or parameters\n",NULL);
				return(TCL_ERROR);
				}

			/* get definers */

			msg[0] = 0x1;		/* message # 1 */
			msg[1] = 0x80;		/* parameter code */
			for(n=2,i=2;i<argc && n < 254;i+=2) {
				id = atoi(argv[i]);
				len = strlen(argv[i+1]);

				if (id < 1 || id > 33) {
					Tcl_AppendResult(interp, "adsi define error: 1 <= ",
							argv[i]," <= 33\n",NULL);
					return(TCL_ERROR);
					}
				if (len > 16) {
					Tcl_AppendResult(interp,"adsi define message (",argv[i],
							") > 16 bytes\n",NULL);
					return(TCL_ERROR);
					}
				msg[n++] = len+1;
				msg[n++] = id;
				strcpy(&msg[n],argv[i+1]);		/* message */
				n += len;
				}

			if (n > 254) {
				sprintf(msg,"%d",n);
				Tcl_AppendResult(interp,"adsi define message ",msg," > 254 chars",
							NULL);
				return(TCL_ERROR);
				}
			n = send_msg(msg,n,S(audio_fd),S(phone_fd));
			break;

		case 2:		/* assign */
			len = argc-2;
			if (len < 1 || len > 8) {
				Tcl_AppendResult(interp,"adsi assign: needs 1-8 parameters\n",
							NULL);
				return(TCL_ERROR);
				}
			msg[0] = 0x1;		/* message # */
			msg[1] = 0x81;		/* function code */
			msg[2] = len;
			for(i=2;i<argc;i++) {
				id = atoi(argv[i]);
				if (id < 1 || id > 33) {
					Tcl_AppendResult(interp,"adsi assign: needs 1-8 parameters\n",
								NULL);
					return(TCL_ERROR);
					}
				msg[i+1] = id;
				}
			n = send_msg(msg,len+3,S(audio_fd),S(phone_fd));
			break;
		}
	sprintf(msg,"%d",n);
	Tcl_SetResult(interp,msg,TCL_VOLATILE);
	return(TCL_OK);
   }

#define MAXDUR		8000		/* 1 seconds */
static char tone_buff[MAXDUR+ MAXDUR/2];
static char quiet[MAXDUR];

int
play_tones(fd,tones,vol,dur)
int fd;
char *tones;
char *vol;
int dur;		/* duration (samples) */
	{
	static int init = 1;
	char c;
	int n = 0;
	int gain = 80;
	if (dur>=MAXDUR)	
		dur = MAXDUR;

	if (vol)
		gain = atoi(vol);

	if (init) {
		int i = MAXDUR;
		while(i>0) quiet[--i] = 127;
		init = 0;
		}

	while (c = *tones++) {
		if (c == '-') {
			int i;
			for(i=0;i<5;i++)
				audio_write(fd,quiet,dur);
			continue;
			}
		if (!gen_tone(c,dur,gain,tone_buff))
			continue;
		audio_write(fd,tone_buff,dur);
		audio_write(fd,quiet,dur/2);
		n++;
		}
	audio_drain(fd);
	return(n);
	}
