/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* simple database routines - PhoneScript interface (S A Uhler) */

#include <stdio.h>
#include <string.h>
#include "tcl.h"
#include "defs.h"

#define dprintf		if (debug) fprintf
#define Dprintf		if (debug>1) fprintf

/* selection critereon */

#define DB_NONE	0
#define DB_SET		1
#define DB_AND		2
#define DB_OR		3

struct db {
	struct data *top;			/* pointer to attr names above 1st row */
	struct data *current;	/* pntr to current item */
	char *name;					/* name of file */
	int state;					/* unused state flags */
	int count;					/* number of lines */
	int fields;					/* number of fields */
	char **item_names;		/* list of item names */
	};

struct data {
	struct data *next;		/* pntr to next row */
	char *data;					/* data in this row */
	char selected;				/* true if item is selected */
	};

int free();
char *malloc(), *getenv(), *index();

static int free_data(struct data *data);	/* !!proto */
static int set_vars(Tcl_Interp *interp, struct db *db);	/* !!proto */
static int count_delim(char *row, int delim);	/* !!proto */

static int debug = 0;

struct db *dbs[20] = {NULL};		/* max # of open db's */

/****************************************************************************/

/* open a database, read into memory, return database cookie */

int
db_set(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db;
	struct data *data, *prev=NULL;
	FILE *file;
	int count = 0;		/* valid line count */
	char *top;
	int line_count = 0;		/* line count */
	int errors=0;
	int i;

	/* see if database is already open, if so make current */

	for(db=dbs[0];db;db++)
		if (strcmp(argv[2],db->name)==0) {
			S(db) = (char *) db;
			dprintf(stderr,"%s: database %s already opened\n",*argv,argv[2]);
			Tcl_SetResult(interp,"2",TCL_VOLATILE);
			return(TCL_OK);
			}

	/* if not, open it, the install it at the end of the list */

	if (!(file=fopen(argv[2],"r"))) {
		Tcl_SetResult(interp,"0",TCL_VOLATILE);
		return(TCL_OK);
		}

	db = (struct db *) malloc(sizeof(struct db));
	db->name = save_line(argv[2]);
	db->state = 0;
	db->fields = 0;
	db->current = NULL;
	S(db) = (char *) db;

	while(fgets(line,MAXLINE,file)) {
		line_count++;
		if (*line=='\n' || (count==0 && *line=='#'))  {	
			Dprintf(stderr,"db:  Skipping %d: %s",line_count,line);
			continue;
			}

		if (db->fields && (i=count_delim(line,';')) != db->fields) {
			char space[10];
			Dprintf(stderr,"db: Line %d, %d != %d\n",line_count,i,db->fields);
			sprintf(space,"%d: ",line_count);
			Tcl_AppendResult(interp,"db set ",argv[2],
				": Wrong number of fields line ",space,line,NULL);
			errors++;
			continue;
			}

		data = (struct data *) malloc(sizeof(struct data));
		data->data = save_line(line);
		data->selected = 1;
		data->next = NULL;
		if (prev)
			prev->next = data;
		else {
			db->top = data;
			if (!(db->fields = count_delim(db->top->data,';'))) {
				dprintf(stderr,"db: no delimeters, adding 1\n");
				strcat(db->top->data,";");
				db->fields = 1;
				errors++;
				}
			Dprintf(stderr,"db: Setting field count to %d\n",db->fields);
			}
		prev = data;
		count++;
		}
	count--;	/* ??? */

	/* get global info */

	db->count = count;

	db->item_names = (char **) malloc((db->fields+1) * sizeof(char *));
	top = save_line(db->top->data);

	for(i=0;i<db->fields;i++) {
		db->item_names[i] = top;
		if (top = index(top,';'))
			*top++ = '\0';
		Dprintf(stderr,"db: item %d is %s\n",i,top);
		}
	db->item_names[i] = NULL;	/* new */

	dprintf(stderr,"d: opened %s, %d items, %d fields\n",
		db->name,count,db->fields);
	if (errors)
		return(TCL_ERROR);
	else {
		Tcl_SetResult(interp,"1",TCL_VOLATILE);
		return(TCL_OK);
		}
	}

/****************************************************************************/

/* write out the database */

static char buff[100];
static char buff2[100];

int
db_write(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct data *data;
	char *path;
	FILE *file;

	if (argv[2])
		path = argv[2];
	else {
		sprintf(buff,"%s.%d",db->name,getpid());
		path=buff;
		}

	dprintf(stderr,"d: writing %s as %s\n",db->name,path);

	if ((file = fopen(path,"w")) == NULL){
		Tcl_AppendResult(interp,"db write: can't open ",argv[1],"\n",NULL);
		return(TCL_ERROR);
		}

	for(data=db->top;data;data=data->next) {
		if (data->selected)
			fwrite(data->data,1,strlen(data->data),file);
		}

	fclose(file);
	if (!argv[2]) {		/* do commit stuff */
		sprintf(buff2,"%s.old",db->name);
		rename(db->name,buff2);
		link(path,db->name);
		unlink(buff2);
		unlink(path);
		}
	return(TCL_OK);
	}

/****************************************************************************/

/* close a database, free up space */

int
db_close(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct db *d;
	int offset = 0;
	struct data *next,*data;

	dprintf(stderr,"d: closing %s\n",db->name);
	for(data=db->top;data;data=next) {
		next = data->next;
		free_data(data);
		}
	free(db->name);
	free(db->item_names[0]);		/* gets all the field names */
	free(db->item_names);
	free(db);

	/* remove database from slot */

	for(d = dbs[0];d;d++) {
		if (d == db)
			offset=1;
		if (offset)
			*d = *(d+1);
		}

	return(TCL_OK);
	}

/****************************************************************************/

/* goto to the top */

int
db_top(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);

	dprintf(stderr,"d: Setting %s to top\n",db->name);
	db->current = db->top;
	return(TCL_OK);
	}

/****************************************************************************/

/* update a value in the current tuple, look for TCL vars for update */

int
db_update(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	char *value;
	int i;

	if (argv[2]) {
		Tcl_AppendResult(interp,"db update: option ",argv[1]," not implemented\n",0);
		return(TCL_ERROR);
		}

	if (!db->current || db->current == db->top) {
		Tcl_AppendResult(interp,"db update: option ",argv[1]," not implemented\n",0);
		Tcl_AppendResult(interp,"db update, no current values\n",NULL);
		return(TCL_ERROR);
		}

	line[0] = '\0';
	for(i=0;i<db->fields;i++) {
		value = Tcl_GetVar(interp,db->item_names[i],0);
		Dprintf(stderr,"db: Update %s =\n",db->item_names[i]); fflush(stderr);
		if (value) strcat(line,value);
		Dprintf(stderr," %s\n",value);
		strcat(line,";");
		}
	strcat(line,"\n");
		
	free(db->current->data);
	db->current->data = save_line(line);
	return(TCL_OK);
	}

/****************************************************************************/

/* delete the current tuple  - leave current pointing at previous */

int
db_delete(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct data *current;
	struct data *prev;

	dprintf(stderr,"d: deleting in %s\n", db->name);

	if (!(current=db->current) || db->current == db->top) {
		dprintf(stderr,"d: No current item to delete\n");
		Tcl_AppendResult(interp,argv[0]," ",argv[1]," nothing to delete\n",0);
		return(TCL_ERROR);
		}

	for(prev=db->top;prev && prev->next != current;prev=prev->next);

	if (prev->next) {
		prev->next = db->current->next;
		db->count--;
		db->current = prev;
		free_data(current);	
		return(TCL_OK);
		}
	else {
		Tcl_AppendResult(interp,"db delete: deletion error\n",NULL);
		return(TCL_ERROR);
		}
	}
		
/****************************************************************************/

/* insert a new line after the current one, make new item current */

int
db_insert(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct data *data;
	char *value;
	int i;


	for(*line='\0',i=0;db->item_names[i];i++) {
		dprintf(stderr,"inserting %s = ",db->item_names[i]);
		value = Tcl_GetVar(interp,db->item_names[i],0);
		dprintf(stderr, "%s\n",value?value:"\"\"");
		if (value) strcat(line,value);
		strcat(line,";");
		}
	strcat(line,"\n");

	data = (struct data *) malloc(sizeof(struct data));
	if (!db->current)
		db->current=db->top;

	data->next = db->current->next;
	data->data = save_line(line);
	data->selected = 1;
	db->current->next = data;
	db->current = data;
	db->count++;
	return(TCL_OK);
	}
		
/****************************************************************************/

/* make next item current, return pointer to row */

int
db_next(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);

	if (db->current)
		db->current = db->current->next;

	while (db->current && db->current->selected==0) {
		db->current = db->current->next;
		}

	if (db->current) {
		set_vars(interp,db);
		dprintf(stderr,"d-next:  (%.10s...)\n",db->current->data);
		Tcl_SetResult(interp,"1",TCL_VOLATILE);
		}
	else {
		set_vars(interp,db);
		dprintf(stderr,"d-next:  NULL\n");
		Tcl_SetResult(interp,"0",TCL_VOLATILE);
		}
	return(TCL_OK);
	}
		
/****************************************************************************/

/* process each tuple in a database (is this redundant with select?) */

int
db_process(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct data *data;
	char *bad;
	int n;
	int i;
	char num[10];

	if ((n=db->count) ==0) {
		return(TCL_OK);
		}

	for(i=1,data=db->top->next;data;data=db->current->next,i++) {
		db->current = data;
		if (!data->selected)
			continue;
		set_vars(interp,db);
		/* Tcl_SetVar(interp,"selected",data->selected ? "1" : "0" ,0); */
		sprintf(num,"%d",i);
		Tcl_SetVar(interp,"NR",num,0);

		if ((n=Tcl_Eval(interp,argv[2],0,&bad)) != TCL_OK) {
			if (n == TCL_CONTINUE) continue;
			if (n == TCL_BREAK) break;
			Tcl_AddErrorInfo(interp,"db process failure:\n\terror at: ");
			error_cmd(interp,argv[2],bad);
			return(TCL_ERROR);
			}
		}
	return(TCL_OK);
	}

/****************************************************************************/

/* return the list of database fields */

int
db_fields(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	int i;

	Tcl_ResetResult(interp);
	for(i=0;db->item_names[i];i++) {
		Tcl_AppendElement(interp,db->item_names[i],0);
		Dprintf(stderr,"db: Fields: %s\n",interp->result);
		}
	return(TCL_OK);
	}

/****************************************************************************/

/* select items from a database */

int
db_select(interp,stuff,argv)
Tcl_Interp *interp;
struct stuff *stuff;
char *argv[];
	{
	struct db *db = (struct db *) S(db);
	struct data *data;		/* current data item */
	int how = DB_NONE;		/* selection mode */
	char *expr;					/* expression to evaluate */
	int result;					/* result of expression */
	int count = 0;				/* # of selected database items */

	/* get select mode and expression */

	if (strcmp(argv[2],"and") == 0)
		how = DB_AND;
	else if (strcmp(argv[2],"or") == 0)
		how = DB_OR;
	else if (strcmp(argv[2],"set") == 0)
		how = DB_SET;

	if (how == DB_NONE) {
		how = DB_SET;
		expr = argv[2];
		}
	else
		expr = argv[3];

	/* loop through items */

	for(data=db->top->next;data;data=data->next) {
		Dprintf(stderr,"db: Select: working on %s",data->data);
		if (data->selected && (how == DB_OR))  {	/* already selected */
			count++;
			Dprintf(stderr,"db: %d) Already Selected (using OR)\n",count);
			continue;
			}
		if (!data->selected && (how == DB_AND)) {	/* already UNselected */
			Dprintf(stderr,"db: %d)  Already DEselected (using AND)\n",count);
			continue;
			}

		/* set up the TCL environment */

		db->current = data;
		set_vars(interp,db);
		Tcl_SetVar(interp,"selected",data->selected ? "1" : "0" ,0);

		/* evaluate the expression */

		if (Tcl_ExprBoolean(interp,expr,&result) != TCL_OK) {
			dprintf(stderr," select expr = %s\n",expr);
			Tcl_AddErrorInfo(interp,"select expression failed");
			return(TCL_ERROR);
			}

		Dprintf(stderr,"db:  %d) %sSelected\n",count,result?"":"UN-");
		if (result) {
			count++;
			data -> selected = 1;
			}
		else
			data -> selected = 0;
		}
	sprintf(line,"%d",count);
	Tcl_SetResult(interp,line,TCL_VOLATILE);
	dprintf(stderr,"d: db_select selected %d\n",count);
	sprintf(line,"%d",count);
	Tcl_SetVar(interp,"select_count",line,0);
	if (count==0)
		set_vars(interp,db);
	return(TCL_OK);
	}

/* free an item */

static int
free_data(data)
struct data *data;
	{
	if (data->data)
		free(data->data);
	return(free(data));
	}

/* allocate space for and save a line (whimpy version) */

char *
save_line(s)
char *s;		/* line to save */
	{
	Dprintf(stderr,"D: saving [%s]\n",s);
	return(strcpy(malloc(strlen(s)+1),s));
	}

/* count the DELIMS in a line */

static int
count_delim(row,delim)
char *row;
char delim;
	{
	int n;

	for(n=0;row&&(row=index(row,delim));n++,row++);
	return(n);
	}
		
void
debug_db(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"db_debug in %d\n",n);
	}

/* set the TCL variable to the current values */

static int
set_vars(interp,db)
Tcl_Interp *interp;
struct db *db;
	{
	int i;
	char *end, *data;
	
	if (!db->current || db->top==db->current) {
		for(i=0;db->item_names[i];i++) 
			Tcl_SetVar(interp,db->item_names[i],"",0);
		dprintf(stderr,"set_vars: no current item\n");
		return(0);
		}

	data = db->current->data;

	for(i=0;db->item_names[i];i++) {
		end = index(data,';');
		*end = '\0';
/*
		Dprintf(stderr,"db: setting TCL var %s = %s\n",db->item_names[i],data);
*/
		Tcl_SetVar(interp,db->item_names[i],data,0);
		*end = ';';
		data = end+1;
		}
	return(1);
	}
