/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* sample TCL intepreter for Phone applications (S. A. Uhler) */

#include <stdio.h>
#include <sys/signal.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <fcntl.h>
#include "tcl.h"
#include "audio_lib.h"
#include "phone_lib.h"
#include "defs.h"
#include "version.h"

#define CID_DEV	"/dev/cid"	/* port for caller id device */
#define CALL		"/dev/call"
#define ANSWER		"/dev/answer"
/*
#define EVAL(in,cmd,flg,bad)	\
	(interactive ? Tcl_RecordAndEval(in, cmd, flg, bad) : \
		Tcl_Eval(in, cmd, flg, bad))
*/
#define EVAL(in,cmd,flg,bad)	Tcl_Eval(in, cmd, flg, bad)

/* make variables global */

#ifndef TCL_GLOBAL_ALWAYS
# define TCL_GLOBAL_ALWAYS	0
#endif
#define GLOBAL		(TCL_GLOBAL_ONLY | TCL_GLOBAL_ALWAYS)

/* these control the audio library parameters */

extern int quiet_thres; 		/* answer quiet threshold */
extern int record_gain;
extern int play_gain;

/* unix externals */

extern int errno;
extern char **environ;			/* UNIX environment - exported to TCL */
	
struct stuff stuff_data, *stuff=&stuff_data;		/* all globals should be here */
char line[MAXLINE];				/* generic buffer */
static int got_sig = 0;			/* got USR1 signal */

static void sig(int n);			/* !!proto */
static int do_on(Tcl_Interp *interp,char *name);	/* !!proto */
static int sethup(char *arg);	/* !!proto */
static int remove_files(Tcl_Interp *interp,char *list);	/* !!proto */
static char *do_debug(ClientData data, Tcl_Interp *interp, char *name1, char *name2, int flags); /* !!proto */

char *cid_read(int fd, char *line);	/* !!proto */
char *getenv(char *var);	/* !!proto */
char *index(char *s, char c);	/* !!proto */
int do_hangup(Tcl_Interp *interp, struct stuff *stuff);	/* !!proto */

/* variables exported to TCL that need inital values */

struct vars vars[] = {
	"calls",				"0",		/* calls processed so far */
	"number",			"",		/* called number (for caller id) */
	"caller_id",		"0",		/* enable caller id */
	"rings",				"1",		/* answer after this many rings */
	"timelimit",	"30", 	/* default timeout on play */
	"record_timeout",	"300",	/* default timeout on record */
	"silence",			"20",		/* default timeout on silence */
	"speed",				"70",		/* orator speed */
	"dial_tone",		"80",		/* time to wait for dial tone (call()) */
	"wait_rings",		"8",		/* # of rings to wait for answer (call()) */
	"sharp",				"",		/* play tone upon answering (usually a "#") */
	"type_ahead",		"1",		/* allow touch-tone type-ahead */
	"reason",			"t",		/* run play unless clause on touch tones */
	"truncate",			"1600",	/* truncate end of recorded message if tone */
	"leave",				"1000",	/* leave # bytes in Q when play is done */
	"tone_duration",	"1200",	/* default phone tones duration 150ms */
	NULL, NULL
	};

static Tcl_Interp *interp;				/* TCL intepreter (should be in stuff??) */
static Tcl_CmdBuf buffer;				/* TCL command buffer */

int
main(argc,argv)
int argc;
char **argv;
	{
	char *call = CALL;				/* outgoing phone device */
	char *answer = ANSWER;			/* incoming phone device */
	char *cid_dev = CID_DEV;		/* name of caller id port */
	struct vars *var = vars;		/* local variable pointer */
	int ring,rings;					/* current ring, answer rings */
	char *args;							/* argument pointer */
	char *oncall;						/* TCL program run for each call */
	char **env = environ;			/* The environment, exported to TCL */
	char *bad;							/* pointer to misteak in eval cmd */
	char *id;							/* caller id info */
	char *cmd="";					/* current TCL command start with dummy */
	int interactive = 0;				/* run interactively */
	FILE *in;							/* input file */
	int caller_id=0;					/* fd of caller id */
	int prompt=0;						/* prompt count for interactive use */
	int no_open = 0;					/* don't ope audio or phone */
	int i;

	S(debug_list) = getenv("DEBUG_LIST");

	if (argc<2) {
		fprintf(stderr,"usage: %s <file>\n",*argv);
		exit(1);
		}

/*
   if (phone_check(call,"hdelp") <0) {
      fprintf(stderr, "Can't set up phone for outgoing calls\n");
      exit(2);
      }
*/

	/* set up TCL */

	signal(SIGINT,term);
	signal(SIGUSR1,SIG_IGN);

   interp = Tcl_CreateInterp();
	buffer = Tcl_CreateCmdBuf();
	init_commands(interp);
	Tcl_TraceVar(interp, "debug", TCL_TRACE_WRITES, do_debug, NULL);
	S(calls) = 0;
	S(phone_fd) = -1;
	S(lineno)=0;
	S(hangup)=0;
	S(resume)=0;
	
	/* set local variables */

	for(;var->name;var++) {
		Tcl_SetVar(interp,var->name,var->value,GLOBAL);
		dprintf("I","Setting [%s] = [%s]\n",var->name,var->value);
		}

	/* export the environment */

	for(;env && (args = *env);env++) {
		char *s = index(args,'=');
		if (s) {
			*s = '\0';	
			Tcl_SetVar(interp,args,s+1,GLOBAL);
			dprintf("I","Setting [%s] = [%s]\n",args,s+1);
			*s = '=';
			}
		else
			Tcl_SetVar(interp,args,"",GLOBAL);
		}

	/* export argc and argv */

	args = Tcl_Merge(argc,argv);	
	Tcl_SetVar(interp,"args",args,GLOBAL);
	dprintf("I","Setting argc = [%d]\n",argc);
	dprintf("I","Setting argv = [%s]\n",args);
	free(args);
	sprintf(line,"%d",argc-1);
	Tcl_SetVar(interp,"argc",line,GLOBAL);
	for(i=1;i<argc;i++) {
		sprintf(line,"%d",i-1);
		Tcl_SetVar2(interp,"argv",line,argv[i],GLOBAL);
		}

	/* other stuff - process id, version */

	sprintf(line,"%d",getpid());
	Tcl_SetVar(interp,"pid",line,GLOBAL);
	Tcl_SetVar(interp,"version",VERSION,GLOBAL);

	/* set the environment variables */

	scale_tones(-3);		/* make touch tones softer */
	if (getenv("QUIET_THRES"))
		quiet_thres = atoi(getenv("QUIET_THRES"));
	if (getenv("PLAY_GAIN"))
		play_gain = atoi(getenv("PLAY_GAIN"));
	if (getenv("RECORD_GAIN"))
		record_gain = atoi(getenv("RECORD_GAIN"));
	if (getenv("ANSWER"))
		answer = getenv("ANSWER");
	if (getenv("CALL"))
		call = getenv("CALL");
	if (getenv("NO_OPEN"))
		no_open = (int) getenv("NO_OPEN");
	if (getenv("CID_DEV"))
		cid_dev = getenv("CID_DEV");

	/* open the phone for outgoing calls */

	if (!no_open) {
		while ((S(phone_fd) = phone_open(call,PH_NODELAY,"AdhElp")) <0) {
			if (errno != EINTR) {
				perror(call);
				term(101);
				}
			}
		}

	/* run initialization/origination pass */

	S(or) = NULL;
	if (!no_open)
		S(audio_fd) = audio_open(2);
	set_callback(sethup,NULL);

	/* See if we need to run interactively */

	if (strcmp(argv[1],"-")==0) {
		interactive = isatty(0);
		in = stdin;
		}
	else if ((in=fopen(argv[1],"r")) == NULL) {
		fprintf(stderr,"%s: Can't open file %s\n",argv[0],argv[1]);
		term(106);
		}

	/* main processing loop */

	while(1) {
		if (S(hangup)) {
			S(hangup) = 0;
			do_hangup(interp,stuff);
			}
		if (interactive) {
			printf("%d%s ",prompt,cmd?(prompt++,"%"):">");
			fflush(stdout);
			}
		if (!fgets(line,sizeof(line),in))
			break;
		dprintf("L","%d: %s",S(lineno),line);
		if ((cmd = Tcl_AssembleCmd(buffer, line)) &&
							 (EVAL(interp, cmd, 0, &bad) != TCL_OK)) {
			error_cmd(interp,cmd,bad);
			if (!interactive) {
				printf("%s\n",interp->result);
				if (S(debug_list))
					Tcl_Eval(interp,"debug",0,&bad);
				else
					do_on(interp,"error");
				term(100);
				}
			}
		if (cmd && interactive)
			printf("%s\n",interp->result);
		S(lineno)++;
		}

	/* check for caller id initialization */

	id = Tcl_GetVar(interp,"caller_id",0);
	if (id && (*id=='1' || *id=='y')) {
		caller_id = cid_open(cid_dev);
		dprintf("i", "opened caller id interface, code %d\n",caller_id);
		if (caller_id <0) {
			fprintf(stderr,"Can't open caller-id\n");
			caller_id = 0;
			}
		}
	else
		caller_id = 0;

	/* run per-call program */

	oncall=Tcl_GetVar2(interp, "on", "call",0);

   while(oncall) {
		audio_close(S(audio_fd));
		phone_flush(S(phone_fd));
		loop_off(S(phone_fd));
		dtmf_on(S(phone_fd));
      phone_close(S(phone_fd));
		S(phone_fd) = -1;

		/* wait here for phone call to come in, process signals as appropriate */

		dprintf("S","waiting for a phone call\n");
		signal(SIGUSR1,sig);
      while ((S(phone_fd) = phone_open(answer,0,0)) <0) {
			if (got_sig) {
				do_on(interp,"signal");
				got_sig = 0;
				}
			else if (errno != EINTR) {
				perror(answer);
				term(102);
				}
			}
		signal(SIGUSR1,SIG_IGN);

		/* see if extension pick up or ringing */

		*line = '\0';
		phone_read(S(phone_fd),line,3,3);
		dprintf("S","phone call type %c\n",*line);

		if (*line == 'H') {	/* extension stuff (figure out what to do later) */
			do_on(interp,"extension");
			dcd_off(S(phone_fd));
			continue;
			}
		if (*line != 'R') {	/* what the??? */
			fprintf(stderr,"Warning, got [%c] <0x%x>, expected 'R'\n",*line,*line);
			}

		/* get caller id stuff */

		rings = atoi(Tcl_GetVar(interp,"rings",0));
		if (caller_id && rings>0) {
			char *number;

			dprintf("i","Looking for caller id\n");
			if (number = cid_read(caller_id,line))
				Tcl_SetVar(interp,"number",number,GLOBAL);
			else
				Tcl_SetVar(interp,"number","",GLOBAL);
			dprintf("i","caller id got [%s]\n",line);
			}

		/* run "ringing" script */

		S(audio_fd) = audio_open(2);
		do_on(interp,"ringing");

		/* wait for proper # of rings */

		rings = atoi(Tcl_GetVar(interp,"rings",0));
		dprintf("m","Waiting for %d rings\n",rings);
		if ((ring=wait_rings(S(phone_fd),rings)) < rings) {
			sprintf(line,"%d",ring);
			Tcl_SetVar(interp,"last_ring",line,GLOBAL);
			do_on(interp,"endringing");
			dprintf("m","Caller hung up at %d/%d rings\n",ring,rings);
			continue;
			}

		/* answered the phone */

		sprintf(line,"%d",++S(calls));
		Tcl_SetVar(interp,"calls",line,GLOBAL);

		/* run the "oncall" script */

		loop_on(S(phone_fd));
		if (Tcl_Eval(interp,oncall,0,&bad) != TCL_OK && !S(hangup)) {
			error_cmd(interp,oncall,bad);
			fprintf(stderr,"TCL oncall: %s\n",interp->result);
			if (S(debug_list)) Tcl_Eval(interp,"debug",0,&bad);
			}
		S(hangup) = 0;
		do_hangup(interp,stuff);
		}
	term(0);
	return(0);
	}

/* interrupt - clean up and exit  */

void
term(n)
int n;
	{
	char *cmd;

	dprintf(".","got signal %d\n",n);

	/* run the "on int" program on ^c (resume is broken) */

	if (n<100) {
		do_on(interp,"int");
		if (S(resume)) {
			S(resume) = 0;
			return;
			}
		}

	/* remove any orator temporary files */

	if (interp && (cmd = Tcl_GetVar(interp,"temp_names",0)) )
		remove_files(interp,cmd);

	/* clean up and exit */

	phone_close(S(phone_fd));
	S(phone_fd) = -1;
	if (S(or)) or_close(S(or));
#ifndef NOGDBM
	gdbm_closeall();
#endif
	Tcl_DeleteInterp(interp);
	Tcl_DeleteCmdBuf(buffer);
	exit(n);
	}

/* print a message contained in "var" */

int
print_msg(interp,var)
Tcl_Interp *interp;
char *var;
	{
	char *msg = Tcl_GetVar(interp,var,0);
	if (msg)
		printf("%s\n",msg);
	return((int)var);
	}

/* called by phone library when a hangup is detected */

static int
sethup(dummy)
char *dummy;
	{
	dprintf("T","Got hangup\n");
	S(hangup)++;
	return(1);
	}

/* set debugging modes: called as WRITE trace on TCL variable debug */

static char *
do_debug(data, interp, name1, name2, flags)
ClientData data;
Tcl_Interp *interp;
char *name1, *name2;
int flags;
	{
	S(debug_list) = Tcl_GetVar(interp,name1,0);
	if (*S(debug_list)) {
		fprintf(stderr,"Setting debug list to [%s]\n",S(debug_list));
		if (index(S(debug_list),'d'))
			debug_db(index(S(debug_list),'D')?2:1);
		if (index(S(debug_list),'s'))
			debug_phone(index(S(debug_list),'S')?2:1);
		}
	return(NULL);
	}

/* remove a list of files */

static int
remove_files(interp,files)
Tcl_Interp *interp;
char *files;				/* list of files to unlink */
	{
	char **list;
	int count;

	dprintf("T","Removing [%s]\n",files);
	if (Tcl_SplitList(interp,files,&count, &list) == TCL_OK) {
		while(count-- > 0) {
			unlink(list[count]);
			}
		free(list);
		return(1);
		}
	return(0);
	}

/* signals USR1 (and USR2) get trapped here */

static void
sig(n)
int n;
	{
	dprintf("S","Got signal %s\n",n);
	got_sig = n;
	}

/* run an "on" command, return 1 iff OK, 0 otherwise */

static int
do_on(interp,name)
Tcl_Interp *interp;
char *name;		/* name of "on" command to do */
	{
	char *prog = Tcl_GetVar2(interp,"on",name,0);
	char *bad;
	int c;

	Tcl_SetVar2(interp,"on","current",name,GLOBAL);
	if (prog) dprintf("S.","Running on command %s\n",name);
	if (prog && (c=Tcl_Eval(interp,prog,0,&bad)) != TCL_OK) {
		error_cmd(interp,prog,bad);
		fprintf(stderr,"Evaluation of %s failed\n",name);
		fprintf(stderr,"%s\n",interp->result);
		if (S(debug_list) && strcmp(name,"int"))
			c = Tcl_Eval(interp,"debug",0,&bad);
		Tcl_SetVar2(interp,"on","current","",GLOBAL);
		return(c);
		}
	Tcl_SetVar2(interp,"on","current","",GLOBAL);
	return(TCL_OK);
	}
