/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* misc PhoneStation subroutines (S. A. Uhler) */

#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include "tcl.h"
#include "defs.h"

/* parse args, stripping off "name=value" constructs
 * return number of args left (preliminary version)
 */

struct vars parms[25];			/* place to hide keyword parameters */

int
parse_args(argv)
char *argv[];
	{
	char *index();
	char *s;					/* the "=" in the keyword */
	char *b;					/* blank pointer */
	int i;					/* the arg # */
	int parm;				/* # of keyword parameters */
	int offset = 0;		/* argv shift counter */
	int count=1;			/* # of positional parameters */

	for(parm=0,i=1;argv[i];i++) {
		dprintf("A","Parse: %d = [%s]\n",i,argv[i]);
		s=index(argv[i],'='); 
		b=index(argv[i],' '); 
		if (s &&	!b &&						/* has an '=' but no blanks */
				s > argv[i]  &&			/* not first character */
				s[-1] && isalnum(s[-1]) &&	/* alpha-numeric in front */
				s[1])  { 	/* something in back */
			*s = '\0';
			parms[parm].name = argv[i];
			parms[parm].value = s+1;
			dprintf("P","  KW: %s = %s\n",argv[i],s+1);
			offset++;
			parm++;
			}
		else {
			argv[i-offset] = argv[i];
			count++;
			}
		}
	argv[i-offset] = argv[i];
	parms[parm].name = NULL;
	return(count);
	}

/* get the value for an optional parameter */

static char *
get_param(var)
char *var;
	{
	int i;

	for(i=0;parms[i].name;i++)
		if (strcmp(var,parms[i].name)==0) {
			dprintf("g","Got local value for %s is [%s]\n",var,parms[i].value);
			return(parms[i].value);
			}
	return(NULL);	
	}

/* look up a value for the variable */

char *
get_value(interp,name)
Tcl_Interp *interp;
char *name;
	{
	char *result;

	if (!(result=get_param(name)))
		result = Tcl_GetVar(interp,name,0);
	dprintf("g","Get value of %s is [%s]\n",name,result?result:"(NONE)");
	return(result?result:"");
	}

/* see if keyword value is true
 * return 1 for true, 0 for false, and -1 for unknown
 */

static char *truisms[] = {
	"y", "yes", "on", "t", "true", "1",  NULL };
static char *falsehoods[] = {
	"n", "no", "off", "f", "false", "0", NULL };

int
is_true(str)
char *str;
	{
	char *s = str;
	char **p;
	

	if (!s)
		return(-1);
	while(*s) {
		*s = tolower(*s);
		s++;
		}
	
	for(p = truisms;*p;p++)
		if (strcmp(str,*p)==0) return(1);
	
	for(p = falsehoods;*p;p++)
		if (strcmp(str,*p)==0) return(0);

	return(-1);
	}

#ifdef CC_BOTCH

#include <varargs.h>

/* this is the K&R/SUN version */

void
dprintf(va_alist) va_dcl
	{
	va_list args;
	char *index();
	char *level;
	char *fmt;

	va_start(args);
	level = va_arg(args,char *);
	if (S(debug_list) && (index(S(debug_list),*level) ||
					(level[1] && index(S(debug_list),level[1])) || 
					*level == '*')) {
		fmt = va_arg(args,char *);
		(void) fprintf(stderr,"%.2s: ",level);
		(void) vfprintf(stderr, fmt, args);
		fflush(stderr);
		}
	va_end(args);
	}

#else

#include <stdarg.h>

/* This is the ANSI C version */

void
dprintf(char *level,char *fmt, ...)
	{
	va_list args;
	char *index();

	va_start(args,fmt);
	if (S(debug_list) && index(S(debug_list),*level)) {
		(void) fprintf(stderr,"%c: ",*level);
		(void) vfprintf(stderr, fmt, args);
		fflush(stderr);
		}
	va_end(args);
	}

#endif

/* print TCL codes */

static char *tcl_codes[] = {
	"TCL_OK",
	"TCL_ERROR",
	"TCL_RETURN",
	"TCL_BREAK",
	"TCL_CONTINUE",
	"TCL_UNWIND",
	};

char *
print_code(n)
int n;
	{
	return(tcl_codes[n]);
	}

/* add current line to error message */

int
error_cmd(interp,cmd,bad)
Tcl_Interp *interp;
char *cmd;					/* TCL command */
char *bad;					/* pointer to bad result */
	{
	char space[20];

	sprintf(space," line offset %d\n",interp->errorLine);
	Tcl_AppendResult(interp,space,NULL);
	return(1);
	}
